from pyjuque.Engine.Models.BotModels import Base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from time import time


def get_session(path='sqlite:///'):
    some_engine = create_engine(path, echo=False)
    Base.metadata.create_all(some_engine)
    Session = sessionmaker(bind=some_engine)
    session = Session()
    return session


def timeit(function, text=None, *args):
	''' Used to print the time it takes to run a certain function. '''
	start = time()
	ret = function(*args)
	end = time()
	if text is not False:
		if text is None or text == "":
			text = function.__name__+" took "
		print(text+str(round(end - start, 4))+" s")
	return ret, end - start



import time
from os import getenv
from yaspin import yaspin
from pyjuque.Engine.Models.BotModels import TABotModel, GridBotModel, getSession
from pyjuque.Engine.Database import InitializeDatabaseTaBot, InitializeDatabaseGridBot
from pyjuque.Engine.BotController import BotController
from pyjuque.Engine.GridBotController import GridBotController
from pyjuque.Exchanges.CcxtExchange import CcxtExchange 
from pprint import pprint 
import functools
# bot_config = {
#     'db_url' : 'sqlite:///my_first_strategy.db',
#     'name' : 'bot1',
#     'type' : 'ta', # 'grid' or 'ta'
#     'exchange' : {
#         'name' : 'binance',
#         'params' : {
#             'api_key': '...',
#             'secret' : '...'
#         },
#     },
#     'symbols' : [],
#     'starting_balance' : 0.0005,
#     'quote_asset': 'BTC',
#     'entry_strategy': {
#         'name': 'BBRSIStrategy',
#         'class' : None,
#         'params': {
#             'rsi_len' : 8,
#             'bb_len' : 100,
#             'rsi_ob' : 50,
#             'rsi_os' : 50
#         }
#     },
#     'entry_settings' : {
#         'initial_entry_allocation': 20,
#         'signal_distance': 0.3
#     },
#     'exit_strategy' : None,
#     'exit_settings' : {
#         'take_profit' : 3,
#         'stop_loss_value': 10
#     },
#     'use_logger' : True,
#     'display_status' : True
# }

def defineBot(bot_config):
    for key in ['name', 'symbols', 'exchange']:
        assert key in bot_config.keys(), '{} should be inside the config object'.format(key)
    for key in ['name', 'params']:
        assert key in bot_config['exchange'].keys(), '{} should be inside the exchange config object'.format(key)
    
    accepted_types = ['ta']
    bot_type = 'ta'
    if bot_config.__contains__('type'):
        bot_type = bot_config['type']
    
    assert bot_type in accepted_types, 'The bot\'s type must be one of {}'.format(accepted_types) 
        
    symbols = bot_config['symbols']
    assert len(symbols) > 0, 'You provided an empty symbols list!' + \
        ' It should hold at least one valid symbol.'
    init_symbol = symbols[0]
    quote_asset = symbols[0].split('/')[1]
    for symbol in symbols:
        symbol_quote = symbol.split('/')[1]
        assert quote_asset == symbol_quote, 'All pairs must be trading against the same' + \
            ' asset, but in this case they don\'t: {}, {}'.format(init_symbol, symbol)
    bot_config['quote_asset'] = quote_asset

    if not bot_config.__contains__('db_url'):
        bot_config['db_url'] = 'sqlite:///{}.db'.format(bot_config['name'])

    bot_controller = _defineTaBot(bot_config)
    
    return bot_controller

def _defineTaBot(bot_config):
    session = getSession(bot_config['db_url'])
    bot_name = bot_config['name']
    exchange_name = bot_config['exchange']['name']
    exchange_params = bot_config['exchange']['params']
    exchange = CcxtExchange(exchange_name, exchange_params)
    bot_model = session.query(TABotModel).filter_by(name=bot_name).first()
    
    if bot_model is None:
        print('No bot found by name: {}. Creating...'.format(bot_name))
        InitializeDatabaseTaBot(session, bot_config)
        return _defineTaBot(bot_config)

    # print('Bot model before init bot_controller', bot_model)
    timeframe = '5m'
    if bot_config.__contains__('timeframe'):
        timeframe = bot_config['timeframe']
    bot_controller = BotController(session, bot_model, exchange, None)
    if bot_config.__contains__('display_status'):
        if bot_config['display_status']:
            status_printer = yaspin()
            bot_controller.status_printer = status_printer
    else:
        status_printer = yaspin()
        bot_controller.status_printer = status_printer
    
    if bot_config.__contains__('strategy'):
        found = False
        if bot_config['strategy'].__contains__('custom'):
            if bot_config['strategy']['custom'] == True:
                found = True
                def nothing(self, symbol): return False, None
                entry_function = nothing
                exit_function = nothing
                if bot_config['strategy'].__contains__('entry_function'):
                    if bot_config['strategy']['entry_function'] not in [None, False]:
                        entry_function = bot_config['strategy']['entry_function']
                if bot_config['strategy'].__contains__('exit_function'):
                    if bot_config['strategy']['exit_function'] not in [None, False]:
                        exit_function = bot_config['strategy']['exit_function']

            bot_controller.checkEntryStrategy = functools.partial(entry_function, bot_controller)
            bot_controller.checkExitStrategy = functools.partial(exit_function, bot_controller)

        if not found and bot_config['strategy'].__contains__('class'):
            bot_controller.strategy = bot_config['strategy']['class'](**bot_config['strategy']['params'])
    
    return bot_controller

# def Main():

#     for key in ['db_url', 'name', 'symbols', 'exchange', 'type']:
#         assert key in bot_config.keys(), '{} should be inside the config object'.format(key)
#     for key in ['name', 'params']:
#         assert key in bot_config['exchange'].keys(), '{} should be inside the exchange config object'.format(key)
    
#     session = getSession(bot_config['db_url'])
#     bot_name = bot_config['name']
#     exchange_name = bot_config['exchange']['name']
#     exchange_params = bot_config['exchange']['params']
#     exchange = CcxtExchange(exchange_name, exchange_params)
#     symbols = bot_config['symbols']
#     bot_type = bot_config['type']

#     bot_model = session.query(Bot).filter_by(name=bot_name).first()
#     if bot_model is None:
#         print('No bot found by name: {}. Creating...'.format(bot_name))
#         if bot_type == 'ta':
#             InitializeDatabaseTaBot(session, bot_config)
#         elif bot_ta == 'grid':
#             InitializeDatabaseGridBot(session, bot_config)
#         Main()

#     strategy = None
#     if bot_config.__contains__('entry_strategy'):
#         if bot_config.__contains__('class'):
#             strategy = bot_config['entry_strategy']['class'](**bot_config['entry_strategy']['params'])

#     bot_controller = None
#     if bot_type == 'ta':
#         bot_controller = BotController(session, bot_model, exchange, strategy)
#     elif bot_type == 'grid':
#         bot_controller = GridBotController(name)
#         bot.create(exchange, symbols[0], total_amount, trade_amount, trade_step, total_trades)

#     if bot_config.__contains__('display_status'):
#         if bot_config['display_status']:
#             status_printer = yaspin()
#             bot_controller.status_printer = status_printer

#     while True:
#         try:
#             bot_controller.executeBot()
#         except KeyboardInterrupt:
#             return
#         bot_controller.status_printer.start()
#         left_to_sleep = bot_config['sleep']
#         while left_to_sleep > 0:
#             if bot_controller.status_printer != None:
#                 open_orders = bot_controller.bot_model.getOpenOrders(bot_controller.session)
#                 bot_controller.status_printer.text = 'Open Orders: {} | Checking signals in {} |'.format(
#                     len(open_orders), 
#                     left_to_sleep)
#             time.sleep(1)
#             left_to_sleep -= 1

# if __name__ == '__main__':
#     Main()


import ccxt
import pandas as pd
from pprint import pprint
from decimal import Decimal
from datetime import datetime
import requests 
import json

class FetchHistoricalDataException(Exception):
    pass

class CcxtExchange():

    def __init__(self, exchange_id, params={}):
        self.exchange_id = exchange_id # 'binance'
        self.exchange_class = getattr(ccxt, exchange_id)
        self.ccxt = self.exchange_class(params)
    

    def getOHLCV(self, symbol, interval, limit=1000, start_time=None, cast_to=float):
        """ Converts cctx ohlcv data from list of lists to dataframe. """
        if start_time != None:
            start_time = int(start_time)
        ohlcv = self.ccxt.fetchOHLCV(
            symbol, interval, since=start_time, limit=limit)
        if len(ohlcv) == 0:
            return pd.DataFrame()
        df = pd.DataFrame(ohlcv)
        df.columns = ['time', 'open', 'high', 'low', 'close', 'volume']
        # transform values from strings to floats
        for col in df.columns:
            df[col] = df[col].astype(cast_to)
        df['date'] = pd.to_datetime(df['time'] * 1000000, infer_datetime_format=True)
        return df



    def getOHLCVHistory(self, symbol, interval, limit=None, start_time=None, cast_to=float):
        """ Converts cctx ohlcv data from list of lists to dataframe. """
        time_now = datetime.timestamp(datetime.now()) * 1000
        if limit != None and limit < 1000:
            return self.getOHLCV(symbol, interval, limit=limit, start_time=start_time, cast_to=cast_to)
        df = self.getOHLCV(symbol, interval, limit=1000, start_time=start_time, cast_to=cast_to)
        if df.empty:
            raise FetchHistoricalDataException(
                '{} doesn\'t provide access to this data.'.format(
                    self.exchange_id))
        time_diff = self.ccxt.parse_timeframe(interval) * 1000
        last_time = df.iloc[-1]['time']
        max_limit = limit
        # print(last_time)
        while last_time + time_diff < time_now:
            l = 1000
            if max_limit != None:
                print(max_limit)
                l = max_limit
                if l >= 1000:
                    l = 1000
                    max_limit = max_limit - 1000
            df2 = self.getOHLCV(symbol, interval, limit=l, start_time=last_time, cast_to=cast_to)
            df = df.append(df2, ignore_index = True)
            last_time = df.iloc[-1]['time']
            if l < 1000:
                return df
            # print(last_time)
            # print(df.loc[len(df)-1])
        return df

    def placeOrder(self, symbol, params, test=False): 
        """ 
        Places order on exchange given a dictionary of parameters.
        """
        return self.ccxt.createOrder(symbol, params)


    def placeMarketOrder(self, symbol, side, amount, test=False, custom_id=False):
        """ 
        Places side (buy/sell) market order for amount of symbol.
        """
        args = dict()
        if custom_id:
            args['clientOrderId'] = custom_id
        # try:
        order = self.ccxt.createOrder(
            symbol=symbol, type='market', side=side, 
            amount=amount, price=None, params=args)
        # except:
        #     return None    
        return order


    def placeLimitOrder(self, symbol, side, amount, price, test=False, custom_id=False):
        """ 
        Places side (buy/sell) limit order for amount of symbol at price. 
        """
        args = dict()
        if custom_id:
            args['clientOrderId'] = custom_id
        # try:
        order = self.ccxt.createOrder(
            symbol=symbol, type='limit', side=side, 
            amount=amount, price=price, params=args)
        # except:
        #     return None
        return order


    def placeStopLossMarketOrder(self, symbol, side, amount, price, test=False, custom_id=False):
        """ Places a STOP_LOSS market order for amount of symbol at price. """
        args = dict()
        if custom_id:
            args['clientOrderId'] = custom_id
        ret = dict()
        if self.exchange_id == 'binance':
            args['stopPrice'] = price
            data = self.ccxt.createOrder(
                symbol=symbol, 
                type='stop_loss_limit', 
                side=side, 
                amount=amount, 
                price=price * 0.98, 
                params=args)
            ret = data
            ret['id'] = data['orderId']
        elif self.exchange_id == 'okex':
            args = {
                'instrument_id' : symbol.replace('/', '-'), 
                'order_type':'1', # trigger
                'mode':'1',     # spot
                'side':side, 
                'size':amount, 
                'trigger_price':price, 
                'algo_type':'2'
                }
            data = self.ccxt.spotPostOrderAlgo(args)
            ret = data
            ret['id'] = data['algo_id']
        else:
            raise NotImplementedError('SL order not implemented for {}'.format(self.exchange_id))
        # print('Placing SL Market order for {} returned'.format(self.exchange_id))
        # pprint(data)
        return ret


    def cancelOrder(self, symbol, order_id, is_custom_id=False):
        """ Cancels order given order id """
        params = {}
        if is_custom_id:
            if self.exchange_id == 'kucoin':
                result = self.ccxt.private_get_orders({'status':'active'})
                orders = result['data']['items']
                for order in orders:
                    if order['clientOid'] == order_id:
                        order_id = order['id']
                        break
            else:
                params['clientOrderId'] = order_id
        return self.ccxt.cancelOrder(order_id, symbol, params)


    def cancelAlgoOrder(self, symbol, order_id, order_type='1', is_custom_id=False):
        params = {}
        if is_custom_id:
            params['clientOrderId'] = order_id
        if self.exchange_id == 'binance':
            # try:
            order = self.ccxt.cancelOrder(order_id, symbol, params)
            # except:
            #     return None
        elif self.exchange_id == 'okex':
            # try:
            data = self.ccxt.spotPostCancelBatchAlgos(
                {
                    'instrument_id':'ETH-BTC', 
                    'algo_ids':[order_id], 
                    'order_type':'1'
                })
            order = data
            # except:
            #     return None
        else:
            raise NotImplementedError('Getting algo orders not implemented for {}'.format(self.exchange_id))
        # print('Canceling algo order for {} returned'.format(self.exchange_id))
        # pprint(order)
        return order


    def getOrder(self, symbol, order_id, is_custom_id=False):
        """ Gets order given order id """ 
        params = {}
        # print('Getting Order {}'.format(order_id))
        if is_custom_id:
            # print('custom id')
            if self.exchange_id == 'kucoin':
                # print('Kucoin, custom id')
                result = self.ccxt.private_get_orders({'status':'active'})
                active_orders = result['data']['items']
                result = self.ccxt.private_get_orders()
                done_orders = result['data']['items']
                # print('kucoin active orders len is {}, done_orders {}'.format(len(active_orders), len(done_orders)))
                for order in active_orders:
                    # pprint(order)
                    if order['clientOid'] == order_id:
                        # print('Found!')
                        # pprint(order)
                        order_id = order['id']
                        break
                for order in done_orders:
                      # pprint(order)
                    if order['clientOid'] == order_id:
                        # print('Found!')
                        # pprint(order)
                        order_id = order['id']
                        break
            else:
                params['clientOrderId'] = order_id
        return self.ccxt.fetchOrder(order_id, symbol, params)
    

    def getAlgoOrder(self, symbol, order_id, order_type='1', is_custom_id=False):
        """ 
        Gets an algo order given order id  (Trigger Order, Stop Loss or Take Profit ) 
        """
        params = {}
        if is_custom_id:
            params['clientOrderId'] = order_id
        if self.exchange_id == 'binance':
            order = self.ccxt.fetchOrder(order_id, symbol, params)
        elif self.exchange_id == 'okex':
            data = self.ccxt.spotGetAlgo(
                {
                    'instrument_id' : symbol.replace('/', '-'), 
                    'order_type' : order_type, 
                    'algo_id': order_id
                })
            order = [next(iter(data))][0]
        else:
            raise NotImplementedError('Getting algo orders not implemented for {}'.format(self.exchange_id))
        # print('Getting order for {} returned'.format(self.exchange_id))
        # pprint(data)
        return order


    def updateSQLOrderModel(self, order, order_response, bot):
        """ 
        Updates an order based on it's state on the exchange given by 
        order_response. Should be part of the exchange interface  """
        # print("Order response is")
        # pprint(order_response)
        if order.is_test:
            if order.side == 'buy':
                order.entry_price = order.price
            if order_response['filled'] is not None:
                order.executed_quantity =  Decimal(order_response['filled'])
            if order_response['status'] is not None:
                order.status = order_response['status']
            if order_response['side'] is not None:
                order.side = order_response['side']
        if not order.is_test:
            if order_response['timestamp'] is not None:
                order.timestamp = datetime.fromtimestamp(
                    order_response['timestamp'] / 1000)
            if order_response['price'] is not None:
                order.price = order_response['price']
                if order.side == 'buy':
                    order.entry_price = order_response['price']
            if order_response['amount'] is not None:
                order.original_quantity =  Decimal(order_response['amount'])
            if order_response['filled'] is not None:
                order.executed_quantity =  Decimal(order_response['filled'])
            if order_response['status'] is not None:
                order.status = order_response['status']
            if order_response['side'] is not None:
                order.side = order_response['side']
            if order_response['type'] is not None:
                order.order_type = order_response['type']
        return order


    def getOrderBook(self, symbol, limit):
        if self.exchange_id == 'binance':
            url = 'https://api.binance.com' + '/api/v3/depth' \
                + "?&symbol=" + str(symbol.replace('/', '')) + "&limit=" + str(limit)
            response = requests.get(url)
            payload = json.loads(response.text)
            return payload
        else: 
            raise NotImplementedError
            


"""
Description:
  Creates a Local Order Book and updates it every second via a 
  WebSocket connection to Binance

"""
from os import getenv
import time
import json 
import datetime
import websocket 
import threading
from pprint import pprint
from decimal import Decimal, Context

from pyjuque.Exchanges.Binance import Binance
from pyjuque.Exchanges.CcxtExchange import CcxtExchange

############## GLOBAL VARIABLES ##############

ws = None                             # Websocket Object
exchange = CcxtExchange('binance', {
    'apiKey': getenv('BINANCE_API_KEY'), 
    'secret': getenv('BINANCE_API_SECRET'),
    'timeout': 30000,
    'enableRateLimit': True,
})      
# The exchange
order_book = dict(counter = 0)        # The (per-symbol) order book
order_book_lock = threading.Lock()    # Order book threading lock
order_book_initialized = dict()       # Whether the (per-symbol) local 
                                      #   order book was initialized
buffered_events = dict(counter = 0)   # Buffers events before local 
                                      #   order book initialization
############## FOR COUNTING ##############
buffered_events_count = 0
unbuffered_events_count = 0

def onOpen(ws):
  print('Opened connection')

def onClose(ws):
  print('Closed connection')

def onError(ws, error):
  print('Got an error')
  print(error)

def onMessage(ws, message):
  json_message = json.loads(message)
  # pprint(json_message['data']['s'])
  symbol = json_message['data']['s']
  global order_book_initialized
  if not order_book_initialized[symbol]:
    global order_book_lock, buffered_events, buffered_events_count
    # print("Got event")
    order_book_lock.acquire()
    # print("Buffering event")
    buffered_events_count += 1
    buffered_events[symbol].append(json_message)
    order_book_lock.release()
  else:
    # print("{}s order book already unlocked".format(symbol))
    global order_book, exchange
    if json_message['data']['u'] < order_book[symbol]['lastUpdateId']:
      pass
      # print("Received older event. Discarding it.")
    else:
      new_asks = insertAsks(
        order_book[symbol]['asks'], json_message['data']['a'])
      new_bids = insertBids(
        order_book[symbol]['bids'], json_message['data']['b'])
      order_book[symbol]['lastUpdateId'] = json_message['data']['u']
      order_book['counter'] += 1

      order_book[symbol]['asks'] = new_asks
      order_book[symbol]['bids'] = new_bids

def insertAsks(previous_asks, received_asks):
  """ Inserts multiple new asks in the order book (assumes 
  that the order book AND the new_asks list are sorted)"""

  new_asks = []

  if len(received_asks) < 1:
    return previous_asks
  if len(previous_asks) < 1:
    return received_asks
  
  # print("Prev")
  # pprint(previous_asks)
  # print("Recv")
  # pprint(received_asks)

  # Uses the merge-sort idea of popping the first element in the lists
  # (which should also be the lowest)
  while len(previous_asks) > 0 and len(received_asks) > 0:
    ask = None
    if Decimal(previous_asks[0][0]) < Decimal(received_asks[0][0]):
      ask = previous_asks.pop(0)
      # print('popped from prev')
    elif Decimal(previous_asks[0][0]) > Decimal(received_asks[0][0]):
      # print('popped from recv')
      ask = received_asks.pop(0)
    else:
      # print('equal, popped from both')
      previous_asks.pop(0)
      ask = received_asks.pop(0)
    
    # print(ask)

    if Decimal(ask[1]) > Decimal(0):
      # print("appended")
      new_asks.append(ask)

  # print("After Merge")
  # pprint(new_asks)

  if len(previous_asks) > 0:
    new_asks.extend(previous_asks)
  elif len(received_asks) > 0:
    new_asks.extend(received_asks)
  
  # print("Complete")
  # pprint(new_asks)

  return new_asks

def insertBids(previous_bids, received_bids):
  """ Inserts multiple new bids in the order book (assumes 
  that the order book AND the new_bids list are sorted)"""

  new_bids = []

  while len(previous_bids) > 0 and len(received_bids) > 0:
    bid = None
    if Decimal(previous_bids[0][0]) > Decimal(received_bids[0][0]):
      bid = previous_bids.pop(0)
    elif Decimal(previous_bids[0][0]) < Decimal(received_bids[0][0]):
      bid = received_bids.pop(0)
    else:
      previous_bids.pop(0)
      bid = received_bids.pop(0)
    
    if Decimal(bid[1]) > Decimal(0):
      new_bids.append(bid)

  if len(previous_bids) > 0:
    new_bids.extend(previous_bids)
  elif len(received_bids) > 0:
    new_bids.extend(received_bids)

  return new_bids

class UpdateOrderBookThread(threading.Thread):
  """ Thread that connects to exchange through websocket and updates 
  local order book """
  def __init__(self, name, socket_url, onUpdate=None):
    threading.Thread.__init__(self)
    self.name = name
    self.socket_url = socket_url
    self.onUpdate = onUpdate

  def run(self):
    print("Start Running")
    print(self.socket_url)

    # websocket.enableTrace(True)
    msg = onMessage
    if self.onUpdate != None:
      def onMessageUpdated(ws, message):
        onMessage(ws, message)
        symbol = None
        try:
          json_message = json.loads(message)
          symbol = json_message['data']['s']
        except:
          print("Exception getting symbol from ws message.")
          pass
        self.onUpdate(symbol)
      msg = onMessageUpdated

    global ws

    ws = websocket.WebSocketApp(
      self.socket_url, 
      on_close=onClose, 
      on_error=onError,
      on_message=msg)
    ws.on_open=onOpen
    ws.run_forever()

class CreateOrderBookThread(threading.Thread):
  """ Thread that collects order book data from exchange and 
  initializes local order book """
  def __init__(self, name, exchange, symbols):
    threading.Thread.__init__(self)
    self.name = name
    self.symbols = symbols
    self.exchange = exchange

  def run(self):
    global order_book, buffered_events, unbuffered_events_count
    global order_book_lock, order_book_initialized, buffered_events_count
    time.sleep(2)
    for symbol in self.symbols:
      ob = self.exchange.getOrderBook(symbol, 50)
    #   print('OB IS')
    #   pprint(order_book)
      order_book[symbol] = dict(
        lastUpdateId=ob['lastUpdateId'], 
        bids=ob['bids'], 
        asks=ob['asks'])

      order_book_lock.acquire()
      unbuffered_events_count += len(buffered_events[symbol])
      for i, event in enumerate(buffered_events[symbol]):
        if event['data']['u'] <= order_book[symbol]['lastUpdateId']:
          pass
        else:
          order_book[symbol]['asks'] = insertAsks(
            order_book[symbol]['asks'], event['data']['a'])
          order_book[symbol]['bids'] = insertBids(
            order_book[symbol]['bids'], event['data']['b'])
        order_book[symbol]['lastUpdateId'] = event['data']['u']
      
      order_book_initialized[symbol] = True
      order_book_lock.release()
      print("total: {}, {}: {} unbuffered".format(
        buffered_events_count, symbol, len(buffered_events[symbol])))

class OrderBook():
  def __init__(self, symbols, onUpdate=None, msUpdate=False):

    self.symbols = symbols
    self.onUpdate = onUpdate
    self.msUpdate = msUpdate

  def startOrderBook(self):
    global buffered_events, order_book_initialized, exchange
    msg = "Creating order books holding {} symbols."\
      .format(len(self.symbols))
    print(msg)
    
    for symbol in self.symbols:
      buffered_events[symbol] = []
      order_book_initialized[symbol] = False

    streams = []
    if self.msUpdate:
      streams = [ symbol.replace('/', '').lower()+"@depth@100ms" for symbol in self.symbols ]
    else:
      streams = [ symbol.replace('/', '').lower()+"@depth" for symbol in self.symbols ]
    
    socket_url = "wss://stream.binance.com:9443/stream?streams=" + \
      '/'.join(streams)
    
    update_order_book_thread = UpdateOrderBookThread(
      "UpdateOrderBook", socket_url, self.onUpdate)
    create_order_book_thread = CreateOrderBookThread(
      "CreateOrderBook", exchange, self.symbols)
    
    update_order_book_thread.start()
    create_order_book_thread.start()

  def getOrderBook(self, symbol=None):
    global order_book
    if symbol == None:
      return order_book
    
    if order_book.__contains__(symbol):
      return order_book[symbol]
    
    return None
  
  def stopOrderBook(self):
    global ws
    ws.close()

  def subscribeToSymbol(self, symbol):

    global ws, exchange, order_book_initialized, buffered_events

    speed = "100ms" if self.msUpdate else ""
    
    msg = {
        "method": "SUBSCRIBE",
        "params":
        [
            "{}@depth{}".format(symbol.replace('/', '').lower(), speed)
        ],
        "id": 1
    }

    ws.send(json.dumps(msg))

    buffered_events[symbol] = []
    order_book_initialized[symbol] = False

    create_order_book_thread = CreateOrderBookThread(
      "CreateOrderBook_{}".format(symbol), exchange, [symbol])
    
    create_order_book_thread.start()

  def unsubscribeFromSymbol(self, symbol):
    global ws

    speed = "100ms" if self.msUpdate else ""
    
    msg = {
        "method": "UNSUBSCRIBE",
        "params":
        [
            "{}@depth{}".format(symbol.replace('/', '').lower(), speed)
        ],
        "id": 2
    }

    ws.send(json.dumps(msg))


  def getOrderBookPrice(self, exchange, symbol, side, quantity, is_quote_quantity):
    """ Calculates the average price we would pay / receive per unit of `symbol` 
    if we wanted to trade `quantity` of that `symbol`, based on its order book"""
    global order_book
    initial_quantity = quantity
    symbol_order_book = order_book[symbol]
    order_book_side = symbol_order_book['asks'] \
      if side == 'buy' else symbol_order_book['bids']

    i, orders, price = 0, [], Decimal('0')
    accounted_for_quantity = Decimal('0')

    qtdif = Decimal('1')
    while qtdif > Decimal('0'):
      try:
        order = order_book_side[i]
      except IndexError:
        raise Exception("There are not enough orders in the Order Book.")
      if is_quote_quantity:
        qty = min(order[1] * order[0], quantity - accounted_for_quantity)
      else:
        qty = min(order[1], quantity - accounted_for_quantity)
      
      price += (order[0] * qty)
      accounted_for_quantity += Decimal(qty)
      qtdif = abs(quantity - accounted_for_quantity)
      i += 1


    #   print("Step {}, qty {}, qtdiff {}, price_point {}, price {}".format(i, qty, qtdif, order[0], price))

    # print("Price {}, qty {}".format(price, quantity))
    return price / quantity



from pyjuque.Engine.Models import TABotModel as Bot, PairModel as Pair, \
    EntrySettingsModel as EntrySettings, ExitSettingsModel as ExitSettings, getSession

def InitializeDatabaseTaBot(session, params={}):
    """ Function that initializes the database
    by creating a bot with two pairs. """
    name = 'My Bot'
    symbols = []
    quote_asset = None
    starting_balance = 0.001
    test_run = False
    initial_entry_allocation = 25
    signal_distance = 0.3
    profit_target = 2
    stop_loss_value = 0
    exit_on_signal = False

    if params.__contains__('name'):
        assert type(params['name']) == str
        name = params['name']

    if params.__contains__('symbols'):
        assert type(params['symbols']) == list
        symbols = params['symbols']

    if params.__contains__('quote_asset'):
        assert type(params['quote_asset']) == str
        quote_asset = params['quote_asset']

    if params.__contains__('starting_balance'):
        assert type(params['starting_balance']) in [int, float]
        starting_balance = params['starting_balance']

    if params.__contains__('test_run'):
        assert type(params['test_run']) == bool
        test_run = params['test_run']

    if params.__contains__('entry_settings'):
        assert type(params['entry_settings']) == dict
        entry_settings = params['entry_settings']
        if entry_settings.__contains__('initial_entry_allocation'):
            assert type(entry_settings['initial_entry_allocation']) in [int, float]
            initial_entry_allocation = entry_settings['initial_entry_allocation']

        if entry_settings.__contains__('signal_distance'):
            assert type(entry_settings['signal_distance']) in [int, float]
            signal_distance = entry_settings['signal_distance']

    if params.__contains__('exit_settings'):
        assert type(params['exit_settings']) == dict
        exit_settings = params['exit_settings']
        if exit_settings.__contains__('take_profit'):
            assert type(exit_settings['take_profit']) in [int, float]
            profit_target = exit_settings['take_profit']

        if exit_settings.__contains__('stop_loss_value'):
            assert type(exit_settings['stop_loss_value']) in [int, float]
            stop_loss_value = exit_settings['stop_loss_value']

        if exit_settings.__contains__('exit_on_signal'):
            assert type(exit_settings['exit_on_signal']) == bool
            exit_on_signal = exit_settings['exit_on_signal']


    myobject = Bot(
        name = name,
        quote_asset = quote_asset,
        starting_balance = starting_balance,
        current_balance = starting_balance,
        test_run = test_run
    )

    session.add(myobject)

    entrysets = EntrySettings(
        initial_entry_allocation = initial_entry_allocation,
        signal_distance = signal_distance
        )
    
    exitsets = ExitSettings(
        profit_target = profit_target,      # in %
        stop_loss_value = stop_loss_value,   # in %
        exit_on_signal = exit_on_signal
        )
    myobject.entry_settings = entrysets
    myobject.exit_settings = exitsets
    session.commit()

    for symbol in symbols:
        pair = Pair(
            bot_id = myobject.id,
            symbol = symbol,
            current_order_id = None
        )
        session.add(pair)

    session.commit()


def InitializeDatabaseGridBot(session, params={}):
    """ Function that initializes the database
    by creating a bot with two pairs. """
    name = 'My Bot'
    symbols = []
    quote_asset = 'BTC'
    starting_balance = 0.001
    test_run = False
    initial_entry_allocation = 25
    signal_distance = 0.3
    profit_target = 2
    stop_loss_value = 0
    exit_on_signal = False

    if params.__contains__('name'):
        assert type(params['name']) == str
        name = params['name']

    if params.__contains__('symbols'):
        assert type(params['symbols']) == list
        symbols = params['symbols']

    if params.__contains__('quote_asset'):
        assert type(params['quote_asset']) == str
        quote_asset = params['quote_asset']

    if params.__contains__('starting_balance'):
        assert type(params['starting_balance']) in [int, float]
        starting_balance = params['starting_balance']

    if params.__contains__('test_run'):
        assert type(params['test_run']) == bool
        test_run = params['test_run']

    if params.__contains__('entry_settings'):
        assert type(params['entry_settings']) == dict
        entry_settings = params['entry_settings']
        if entry_settings.__contains__('initial_entry_allocation'):
            assert type(entry_settings['initial_entry_allocation']) in [int, float]
            initial_entry_allocation = entry_settings['initial_entry_allocation']

        if entry_settings.__contains__('signal_distance'):
            assert type(entry_settings['signal_distance']) in [int, float]
            signal_distance = entry_settings['signal_distance']

    if params.__contains__('exit_settings'):
        assert type(params['exit_settings']) == dict
        exit_settings = params['exit_settings']
        if exit_settings.__contains__('take_profit'):
            assert type(exit_settings['take_profit']) in [int, float]
            profit_target = exit_settings['take_profit']

        if exit_settings.__contains__('stop_loss_value'):
            assert type(exit_settings['stop_loss_value']) in [int, float]
            stop_loss_value = exit_settings['stop_loss_value']

        if exit_settings.__contains__('exit_on_signal'):
            assert type(exit_settings['exit_on_signal']) == bool
            exit_on_signal = exit_settings['exit_on_signal']


    myobject = Bot(
        name = name,
        quote_asset = quote_asset,
        starting_balance = starting_balance,
        current_balance = starting_balance,
        test_run = test_run
    )

    session.add(myobject)

    entrysets = EntrySettings(
        initial_entry_allocation = initial_entry_allocation,
        signal_distance = signal_distance
        )
    
    exitsets = ExitSettings(
        profit_target = profit_target,      # in %
        stop_loss_value = stop_loss_value,   # in %
        exit_on_signal = exit_on_signal
        )
    myobject.entry_settings = entrysets
    myobject.exit_settings = exitsets
    session.commit()

    for symbol in symbols:
        pair = Pair(
            bot_id = myobject.id,
            symbol = symbol,
            current_order_id = None
        )
        session.add(pair)

    session.commit()




import time
import math
from uuid import uuid4
from sys import exc_info
from pprint import pprint
from traceback import print_exc
from pyjuque.Engine.Models import TABotModel as Bot, PairModel as Pair, OrderModel as Order


def placeNewOrder(exchange, symbol, pair=None, order=None, test_mode=True, order_params=None):
    """ Create Order model and place order to exchange. """
    new_order_model = createOrderModel(symbol, test_mode, order_params, order)
    if not test_mode:
        new_order_response = placeOrderFromOrderModel(exchange, new_order_model)
    else:
        new_order_response = dict(message='success', status='open', filled=0, side=new_order_model.side)
    exchange.updateSQLOrderModel(new_order_model, new_order_response, None)
    return new_order_model


def createOrderModel(symbol, test_mode, order_params, order):
    """ Create Order Model and fill only mandatory params. 
    Other params are filled after order is filled. """
    if 'price' not in order_params:
        order_params['price'] = None
    if 'take_profit_price' not in order_params:
        order_params['take_profit_price'] = None
    if 'stop_price' not in order_params:
        order_params['stop_price'] = None
    if 'is_entry' not in order_params:
        order_params['is_entry'] = False
    if order is None:
        position_id = str(uuid4()).replace('-', '')
        entry_price = None
    elif order is not None:
        position_id = order.position_id
        entry_price = order.entry_price
    new_order_model = Order(
        id = str(uuid4()).replace('-', ''),
        position_id = position_id,
        bot_id = order_params['bot_id'],
        symbol = symbol,
        price = order_params['price'],
        take_profit_price = order_params['take_profit_price'],
        order_type = order_params['order_type'],
        stop_price = order_params['stop_price'],
        original_quantity = order_params['quantity'],
        side = order_params['side'],
        is_entry =  order_params['is_entry'],
        is_test = test_mode,
        entry_price = entry_price,
        last_checked_time = int(round(time.time() * 1000))
    )
    return new_order_model


def placeOrderFromOrderModel(exchange, order_model):
    """ Places orders from db model to exchange."""
    if order_model.order_type == 'limit':
        order_response = exchange.placeLimitOrder(
            order_model.symbol, order_model.side, 
            order_model.original_quantity, order_model.price, 
            order_model.is_test, custom_id=order_model.id)
    if order_model.order_type == 'market':
        order_response = exchange.placeMarketOrder(
            order_model.symbol, order_model.side, 
            order_model.original_quantity, order_model.is_test, 
            custom_id=order_model.id)
    if order_model.order_type == 'stop_loss':
        order_response = exchange.placeStopLossMarketOrder(
            order_model.symbol, order_model.side, 
            order_model.original_quantity, order_model.price, 
            order_model.is_test, custom_id=order_model.id)
    return order_response


def simulateOrderInfo(exchange, order, kline_interval):
    """ Used when BotController is in test mode. 
    Simulates order info returned by exchange."""
    order_status = dict()
    new_last_checked_time = int(round(time.time() * 1000)) # time in ms
    time_diff = new_last_checked_time - order.last_checked_time
    interval_in_ms = exchange.ccxt.parse_timeframe(kline_interval)
    if  time_diff < interval_in_ms:
        candlestick_data = exchange.getOHLCV(order.symbol, kline_interval, 1)
    else:
        minimum_period = int(math.ceil(time_diff / interval_in_ms))
        candlestick_data = exchange.getOHLCV(order.symbol, kline_interval, 
            minimum_period, start_time=order.last_checked_time)

    for _, candle in candlestick_data.iterrows():
        if order.order_type == 'limit':
            if (order.side == 'buy' and candle['low'] < order.price) or \
                (order.side == 'sell' and candle['high'] > order.price):
                order_status['status'] = 'closed'
                order_status['side'] = order.side
                order_status['filled'] = order.original_quantity
                break
        elif order.order_type == 'market':
            order.price = candle['close']
            order_status['status'] = 'closed'
            order_status['side'] = order.side
            order_status['filled'] = order.original_quantity
            break
        elif order.order_type == 'stop_loss':
            if (order.side == 'sell' and candle['low'] < order.price) or \
                (order.side == 'buy' and candle['high'] > order.price):
                order_status['status'] = 'closed'
                order_status['side'] = order.side
                order_status['filled'] = order.original_quantity
                break
    if not order_status:
        order_status['status'] = order.status
        order_status['side'] = order.side
        order_status['filled'] = order.executed_quantity
    order.last_checked_time = new_last_checked_time
    # print('order_status is')
    # pprint(order_status)
    return order_status


def cancelOrder(exchange, order):
    if order.order_type == 'stop_loss':
        return exchange.cancelAlgoOrder(order.symbol, order.id, is_custom_id=True)

    return exchange.cancelOrder(order.symbol, order.id, is_custom_id=True)


# We replaced klineIntervalToMs to exchange.ccxt.parse_timeframe()


"""
    Backtester infrastructure supporting
        - stop loss, trailing stop loss &
        - subsequent entries (DCA) logic

    TODO: Needs proper testing.
"""

import pandas as pd
from decimal import Decimal

# HELPER CLASS
class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

model_entry_strategy:dotdict = dotdict(dict(
    strategy_class = "some_function",
    args = "some_args",
))

model_entry_settings:dotdict = dotdict(dict(
    # subsequent entries
    se = dotdict(dict(
        times = 1,
        after_profit = 0.995,
        pt_decrease = 0.998,
    ))
))

model_exit_settings:dotdict = dotdict(dict(
    profit_target = 1.045,
    # trailing stop loss
    tsl = dotdict(dict(
        value = 0.97,
        after_profit = 1.01
    )),
    # stop loss
    sl = 0.85
))

def backtest(df, symbol, exchange,
    entry_strategy=model_entry_strategy,
    entry_settings=model_entry_settings,
    exit_settings=model_exit_settings):
    '''
        Function used to backtest a strategy on a dataframe `df`
        containing candlestick data of a coin over a period of time.
        Parameters
        --
            DataFrame df - the container of the candlestick data
            dict sd - the dict containing symbol data
            Exchange exchange - exchange to test this strategy on
            dict `entry_strategy` - details about the entry strategy (function & args)
            dict `entry_settings` - details about the entry settings (subsequent entries)
            dict `exit_settings` - details about the exit settings (stop loss, tsl & take profit)
        Returns
        --
            dict information about the backtesting results
    '''

    assert exit_settings.sl is None or \
        (exit_settings.sl <= Decimal(1) and exit_settings.sl > Decimal(0)), \
        ("stop_loss should be between 0 and 1, not "+str(exit_settings.sl))

    assert exit_settings.tsl is None or \
        (exit_settings.tsl.__contains__('value') and \
            exit_settings.tsl.__contains__('after_profit')), \
        ("tsl (exit) settings should contain `value` and `after_profit` ")

    assert exit_settings.tsl is None or \
        exit_settings.tsl.value > Decimal(0) and \
        exit_settings.tsl.value < Decimal(1), \
        "tsl (exit) settings `value` should be between 0 and 1, not "+str(exit_settings.tsl.value)

    assert exit_settings.tsl is None or \
        exit_settings.tsl.after_profit > Decimal(1), \
        ("tsl (exit) settings `after_profit` should be greater than 1, not "\
        +str(exit_settings.tsl.after_profit))

    assert exit_settings.pt is None or \
        exit_settings.pt  > Decimal(1), \
        ("profit_target should be greater than 1, not "+str(exit_settings.pt))

    # lists containing buy and sell times, for plotting
    buy_times = []
    tp_sell_times = []
    sl_sell_times = []
    tsl_sell_times = []
    tsl_active_times = []
    tsl_increase_times = []
    profits_list = []

    buy_price = 0
    subsequent_buys = 0
    resulting_percentage = Decimal(100)

    last_buy = None
    sl_price = None
    tsl_active = False
    tsl_sell_price = None
    next_entry_price = None
    next_target_price = None
    tsl_increase_price = None
    tsl_activate_after = None

    strategy = entry_strategy.strategy_class(*entry_strategy.args)
    strategy.setUp(df)
    # Go through all the candlesticks
    for i in range(0, len(df['close'])-1):
        # Have we already opened a position?
        if last_buy is None:
            # If no, check whether the strategy is fulfilled at this point in time
            strategy_result = strategy.checkLongSignal(i)

            if strategy_result:
                # If strategy is fulfilled, buy the coin
                buy_price = exchange.toValidPrice(symbol, df['close'][i])
                buy_times.append([df['time'][i], buy_price])

                # Initialize TAKE PROFIT PRICE
                if exit_settings.pt is not None:
                    next_target_price = exchange.toValidPrice(symbol,\
                        buy_price * Decimal(exit_settings.pt), round_up=True)

                # Initialize STOP LOSS PRICE
                if exit_settings.sl is not None:
                    sl_price =  exchange.toValidPrice(symbol,\
                        buy_price * Decimal(exit_settings.sl))

                # Initialize TRAILING STOP LOSS PRICE
                if exit_settings.tsl is not None:
                    tsl_activate_after =  exchange.toValidPrice(symbol,\
                        buy_price * Decimal(exit_settings.tsl.after_profit))

                    if tsl_activate_after <= buy_price:
                        tsl_active = True
                        tsl_increase_price = buy_price
                        tsl_sell_price =  exchange.toValidPrice(symbol,\
                            buy_price * Decimal(exit_settings.tsl.value))

                # Initialize SUBSEQUENT ENTRY PRICE
                if entry_settings.se is not None:
                    next_entry_price = exchange.toValidPrice(symbol,\
                        buy_price * Decimal(entry_settings.se.after_profit))

                last_buy = {
                    "index": i,
                    "price": buy_price
                }

        elif last_buy is not None and i > last_buy["index"] + 1:
        # If we already opened a position, check whether the price has hit
        # either the stop loss price, tsl_price, the target price, or if a
        # subsequent entry is due

            ### TRAILING STOP LOSS LOGIC
            if exit_settings.tsl is not None:
                if not tsl_active:
                    if tsl_activate_after <= Decimal(df['high'][i]):
                        tsl_active_times.append((df['time'][i], tsl_activate_after))
                        tsl_active = True
                        tsl_increase_price = Decimal(df['high'][i])
                        tsl_sell_price = exchange.toValidPrice(symbol,\
                            Decimal(df['high'][i]) * Decimal(exit_settings.tsl.value))
                if tsl_active:
                    if Decimal(df['low'][i]) <= tsl_sell_price:
                    # Price went below TSL so we have to sell
                        profits_list.append(tsl_sell_price - last_buy['price'])
                        tsl_sell_times.append([df['time'][i], tsl_sell_price, subsequent_buys])
                        resulting_percentage = resulting_percentage * (tsl_sell_price / buy_price)
                        buy_price = Decimal(0)
                        tsl_active = False
                        tsl_activate_after = None
                        tsl_increase_price = None
                        last_buy = None
                        subsequent_buys = 0
                    elif Decimal(df['high'][i]) > tsl_increase_price:
                    # Price went above pervious high so we adjust TSL Target
                        tsl_increase_times.append((df['time'][i], tsl_increase_price))
                        tsl_increase_price = Decimal(df['high'][i])
                        tsl_sell_price = exchange.toValidPrice(symbol,\
                            Decimal(df['high'][i]) * Decimal(exit_settings.tsl.value))

            ### STOP LOSS LOGIC
            if exit_settings.sl is not None and \
                Decimal(df['low'][i]) < sl_price:
                # If price went below our stop_loss, it means we sold at that point
                profits_list.append(sl_price - last_buy['price'])
                sl_sell_times.append([df['time'][i], sl_price, subsequent_buys])
                resulting_percentage = resulting_percentage * (sl_price / buy_price)
                buy_price = Decimal(0)
                subsequent_buys = 0
                tsl_active = False
                tsl_activate_after = None
                tsl_increase_price = None
                last_buy = None

            ### SUBSEQUENT ENTRIES LOGIC
            if entry_settings.se is not None and \
                Decimal(df['low'][i]) < next_entry_price \
                and subsequent_buys < entry_settings.se.times:

                buy_price = next_entry_price

                if entry_settings.pt is not None:
                    next_target_price = exchange.toValidPrice(symbol,\
                        buy_price * Decimal(entry_settings.pt) * \
                        Decimal(entry_settings.se.pt_decrease), round_up=True)

                if exit_settings.tsl is not None:
                    tsl_activate_after = exchange.toValidPrice(symbol,\
                        buy_price * Decimal(exit_settings.tsl.after_profit))

                    if tsl_activate_after <= buy_price:
                        tsl_active = True
                        tsl_increase_price = buy_price
                        tsl_sell_price =  exchange.toValidPrice(symbol,\
                            buy_price * Decimal(exit_settings.tsl.value))
                    else:
                        tsl_active = False

                next_entry_price = exchange.toValidPrice(symbol,\
                    buy_price * Decimal(entry_settings.se.after_profit))
                
                buy_times.append([df['time'][i], buy_price])
                last_buy = { "index": i, "price": buy_price}
                subsequent_buys = subsequent_buys + 1

            ### TAKE PROFIT LOGIC
            if exit_settings.pt is not None and exit_settings.tsl is None and \
                Decimal(df['high'][i]) > next_target_price:

                tp_sell_times.append([df['time'][i], next_target_price, subsequent_buys])
                profits_list.append(next_target_price - last_buy['price'])
                resulting_percentage = resulting_percentage * \
                    (next_target_price / buy_price)
                tsl_increase_price = None
                last_buy = None
                subsequent_buys = 0
                tsl_active = False
                tsl_activate_after = None
                buy_price = Decimal(0)

    ms = df['time'][len(df['time'])-1] - df['time'][0]
    return dict(
        total_profit_loss = round(float(resulting_percentage), 2),
        buy_times = buy_times,
        tp_sell_times = tp_sell_times,
        sl_sell_times = sl_sell_times,
        tsl_sell_times = tsl_sell_times,
        tsl_active_times = tsl_active_times,
        tsl_increase_times = tsl_increase_times,
        profits_list = profits_list,
        start_time = df['time'][0],
        end_time = df['time'][len(df['time'])-1],
        seconds_of_backtesting = ms/1000,
        days_of_backtesting = round((ms/(1000 * 60 * 60 * 24)), 1)
    )


from pyjuque.Engine.Backtester import backtest
from pyjuque.Engine.BacktesterSundayTheQuant import Backtester
from pyjuque.Engine.Models import *
from pyjuque.Engine.BotController import BotController


'''
-> Open Long/Short position
-> Close Position
-> Set Take Profit
-> Set Stop Loss

'''

import pandas as pd 
import numpy as np 


class Backtester():

    def __init__(self, params = {}):
        self.params = params
        self.initial_balance = 0

        for elem in ['starting_balance', 'entry_settings', 'exit_settings', 'strategy']:
            assert elem in list(params.keys()), 'Key {} should be inside the params dict! '.format(elem)
        
        self.initial_balance = params['starting_balance']
        
        self.leverage = 1
        if params['entry_settings'].__contains__('leverage'):
            self.leverage = params['entry_settings']['leverage']
        self.initial_entry_allocation = 100
        if params['entry_settings'].__contains__('initial_entry_allocation'):
            self.initial_entry_allocation = params['entry_settings']['initial_entry_allocation']

        self.take_profit = 1.05
        if params['exit_settings'].__contains__('take_profit'):
            self.take_profit = (100 + params['exit_settings']['take_profit']) / 100
        self.stop_loss_value = 0
        if params['exit_settings'].__contains__('stop_loss_value'):
            self.stop_loss_value = (100 - params['exit_settings']['stop_loss_value']) / 100
        self.trailing_stop_loss = False
        if params['exit_settings'].__contains__('trailing_stop_loss'):
            self.trailing_stop_loss = params['exit_settings']['trailing_stop_loss']
        self.exit_on_short = False
        if params['exit_settings'].__contains__('exit_on_signal'):
            self.exit_on_short = params['exit_settings']['exit_on_signal']
        self.exit_on_long = False
        if params['exit_settings'].__contains__('exit_on_signal'):
            self.exit_on_long = params['exit_settings']['exit_on_signal']
        self.sell_on_end = False
        if params['exit_settings'].__contains__('sell_on_end'):
            self.sell_on_end = params['exit_settings']['sell_on_end']

        self.strategy = params['strategy']['class'](**params['strategy']['params'])

        self.amount = 0
        self.fee_cost = 0.1 / 100
        self.balance = self.initial_balance
        self.locked_in_trades = 0
        self.locked_trades = 0
        self.first_price = 0
        self.last_price = 0
        self.unrealised_profits = 0
        self.open_positions = 0

        self.inv = self.balance * (self.initial_entry_allocation / 100) * self.leverage

        self.profit = []
        self.drawdown = []

        self.entries = []
        self.exits = []
        self.winned = 0
        self.lossed = 0

        self.num_operations = 0
        self.num_longs = 0
        self.num_shorts = 0

        self.is_long_open = False
        self.is_short_open = False
        self.from_opened = 0


    def reset_results(self):
        self.balance = self.initial_balance
        self.amount = 0
        self.profit = []
        self.drawdown = []
        self.entries = []
        self.exits = []
        self.winned = 0
        self.lossed = 0
        self.num_operations = 0
        self.num_longs = 0
        self.num_shorts = 0
        self.is_long_open = False
        self.is_short_open = False
        self.from_opened = 0
        self.locked_in_trades = 0
        self.locked_trades = 0
        self.last_price = 0
        self.unrealised_profits = 0
        self.open_positions = 0


    def open_position(self, price, time, side, from_opened = 0):
        self.num_operations += 1
        if side == 'long':
            self.num_longs += 1
            # comment
            # if self.is_short_open and self.exit_on_long:
            #     self.close_position(price)
            if self.is_long_open:
                self.long_open_price = (self.long_open_price + price)/2
                self.amount += self.inv #* price
            else:
                self.is_long_open = True
                self.long_open_price = price
                self.amount = self.inv #* price
            self.entries.append([time, price, self.inv, self.inv / price]) #* price ])
        # elif side == 'short':
        #     self.num_shorts += 1
        #     # comment
        #     if self.is_long_open and self.exit_on_short:
        #         self.close_position(price)
        #     elif self.is_short_open:
        #         self.short_open_price = (self.short_open_price + price)/2
        #         self.amount += self.inv #* price
        #     else:
        #         self.is_short_open = True
        #         self.short_open_price = price
        #         self.amount = self.inv #* price
        #     self.entries.append([time, price, self.inv, self.inv / price ]) #* price ])
        self.balance -= self.inv
        self.open_positions += 1
        # self.amount = self.inv/price
        if self.trailing_stop_loss:
            self.from_opened = from_opened


    def close_position(self, price, time):
        if self.is_long_open:
            result = self.amount * (price - self.long_open_price)
            self.is_long_open = False
            self.long_open_price = 0
        else:
            return
        self.num_operations += 1
        # elif self.is_short_open:
        #     result = self.amount * (self.short_open_price - price)
        #     self.is_short_open = False
        #     self.short_open_price = 0
        self.profit.append(result)
        self.exits.append([time, price, self.amount, self.amount / price])
        # print('closing position! {}, {} amount was {}'.format(time, price, self.amount))
        self.balance += self.inv * self.open_positions
        self.open_positions = 0
        self.amount = 0
        if result > 0:
            self.winned += 1
            self.drawdown.append(0)
        else:
            self.lossed += 1
            self.drawdown.append(result)
        self.take_profit_price = 0
        self.stop_loss_price = 0


    def set_take_profit(self, price, tp_long = 1.01, tp_short = 0.99):
        # Here you could pass the tp long/short
        # tp_long = self.params['tp_long']
        # tp_short = self.params['tp_short']
        if self.is_long_open:
            self.take_profit_price = price * tp_long
        # elif self.is_short_open:
        #     self.take_profit_price = price * tp_short


    def set_stop_loss(self, price, sl_long = 0.99, sl_short = 1.01):
        # Here you could pass the sl long/short
        # sl_long = self.params['sl_long']
        # sl_short = self.params['sl_short']
        if self.is_long_open:
            self.stop_loss_price = price * sl_long
        # if self.is_short_open:
        #     self.stop_loss_price = price * sl_short


    def return_results(self):
        # And here also you could pass symbol and period of backtest
        # symbol = self.params['symbol']
        # start_date = self.params['start_date']
        # end_date = self.params['end_date']
        buy_n_hold = self.initial_balance * (self.last_price / self.first_price)
        profit = sum(self.profit) / self.last_price
        drawdown = sum(self.drawdown)
        fees = (abs(profit) * self.fee_cost * self.num_operations) / self.last_price
        winrate = 0
        if (self.winned + self.lossed) > 0:
            winrate = self.winned / (self.winned + self.lossed)

                # PLOT BOT ENTRIES AND EXITS
        entries = []
        exits = []
        # Update sizes of signals on plot
        entry_sizes = [e[2] for e in self.entries]
        exit_sizes = [e[2] for e in self.exits]
        min_size = min(min(entry_sizes), min(exit_sizes))
        max_size = max(max(entry_sizes), max(exit_sizes))
        size_multiplier = 1
        if max_size > 30:
            size_multiplier = 100 / (max_size + min_size)
        for res in self.entries:
            res_clone = res.copy()
            res_clone[2] = res_clone[2] * size_multiplier
            entries.append(res_clone)
        for res in self.exits:
            res_clone = res.copy()
            res_clone[2] = res_clone[2] * size_multiplier
            exits.append(res_clone)
        self.entries = entries
        self.exits = exits

        results = {
            'balance_initial' : self.initial_balance,
            'balance_locked' : self.initial_balance - self.balance,
            'balance_free' : self.balance,
            'balance_plus_profits': self.balance + profit,
            'profit' :	profit + self.unrealised_profits,
            'profit_realised': profit, 
            'profit_unrealised' : self.unrealised_profits,
            'profit_buy_and_hold' : buy_n_hold - self.initial_balance,
            # 'drawdown': drawdown,
            # 'entries': self.entries,
            # 'exits': self.exits,
            'profit_after_fees': profit - fees,
            'n_trades' : self.num_operations,
            'n_long' : self.num_longs,
            'n_shorts': self.num_shorts,
            'n_locked_trades': self.locked_trades,
            'winned' : self.winned,
            'lossed' : self.lossed,
            'winrate' : winrate
        }
        # results['fitness_function'] = ( ( profit - abs(drawdown) ) * winrate) / self.num_operations
        return results


    def backtest(self, df):
        high = df['high']
        close = df['close']
        low = df['low']
        time = df['time']

        self.strategy.setUp(df)

        for i in range(len(df)):

            # Check Signals
            long_signal = self.strategy.checkLongSignal(i)
            short_signal = self.strategy.checkShortSignal(i)
            
            # Close Existing Trades if Open
            if self.is_long_open:
                if self.exit_on_short and short_signal:
                    self.close_position(price = close[i], time = time[i])
                elif high[i] >= self.take_profit_price:
                    self.close_position(price = self.take_profit_price, time = time[i])
                elif low[i] <= self.stop_loss_price:
                    self.close_position(price = self.stop_loss_price, time = time[i])
            # elif self.is_short_open:
            #     if self.exit_on_long and long_signal:
            #         self.close_position(price = close[i], time = time[i])
            #     elif high[i] >= self.stop_loss_price:
            #         self.close_position(price = self.stop_loss_price, time = time[i])
            #     elif low[i] <= self.take_profit_price:
            #         self.close_position(price = self.take_profit_price, time = time[i])
                
            if self.balance > 0:
                # Open New Trades If We Received Signals
                if long_signal:
                    self.open_position(price = close[i], time = time[i], side = 'long', from_opened = i)
                    self.set_take_profit(price = close[i], tp_long = self.take_profit)
                    self.set_stop_loss(price = close[i], sl_long = self.stop_loss_value)
                # elif short_signal:
                #     self.open_position(price = close[i], time = time[i], side = 'short', from_opened = i)
                #     self.set_take_profit(price = close[i], tp_short = self.take_profit)
                #     self.set_stop_loss(price = close[i], sl_short = self.stop_loss_value)
                        
            # Update Trailing Stop Loss If Available
            if self.trailing_stop_loss and (self.is_long_open): # or self.is_short_open):
                new_max = high[self.from_opened : i].max()
                previous_stop_loss = self.stop_loss_price
                self.set_stop_loss(price = new_max)
                if previous_stop_loss > self.stop_loss_price:
                    self.stop_loss_price = previous_stop_loss
                
        if (len(self.entries) > len(self.exits)) and self.is_long_open and self.sell_on_end:
            self.close_position(price = close[i], time = time[i])

        if (len(self.entries) == 0) or (len(self.exits) == 0):
            raise ValueError("""
                Not enough data for computing profits. Either your long or short signal wasn't triggered.
                Please check your strategies and try again. If you have at least one entry but no exits,
                you can force a full sell on the end by setting 'sell_on_end' to True in the exit_settings
                of your bot config. (Total entries: {} / Total exits: {})""".format(
                    len(self.entries), len(self.exits)))

        self.first_price = close[0]
        last_price = close[len(close) - 1]

        remaining_amount = 0
        unrealised_profits = 0
        locked_trades = 0
        if len(self.exits) > 0:
            last_exit = self.exits[len(self.exits) - 1]
            for entry in self.entries:
                if entry[0] > last_exit[0]:
                    remaining_amount += entry[2]
                    unrealised_profits += entry[2] / entry[1]
                    locked_trades += 1
        # self.locked_in_trades = remaining_amount 
        self.last_price = last_price
        self.locked_trades = locked_trades
        self.unrealised_profits = (remaining_amount / last_price - unrealised_profits) #/ last_price
        




import os
import plotly.graph_objs as go
from plotly.offline import plot
import random

"""
    This file contains all the tools used for plotting data.

    It currently contains a general purpose plotting function which plots 
    candlestick charts, overlayed with indicators, signals, & trendlines.

    Eventually, there should be more functions here, each of which would solve a 
    single purpose, to provide more flexibility to developers as to what to plot 
    and how to style it. Modular is the aim.
"""

# TI & TA
def GetPlotData(df, 
    add_volume:bool=True,
    add_candles:bool=True,
    buy_signals:bool or list=False,
    sell_signals:bool or list=False,
    signals:bool or list=False,
    regimes_number=None,
    plot_indicators=[],
    trend_points=False,
    trends=False):
    """ Generates the plotly traces to be plotted. """

    data=[]

    if regimes_number != None:
        if df.__contains__('regime'):
            r_colors = []
            for i in range(regimes_number):
                r, g, b = random.randint(0, 255), random.randint(0, 255), random.randint(0, 255)
                rand_str = 'rgba({}, {}, {}, 110)'.format(r, g, b)
                r_colors.append(rand_str)
                print('Color string for regime {} is {}'.format(i, rand_str))

            last_regime = None
            for i, row in df.iterrows():
                if last_regime == None:
                    last_regime = row['regime']
                    regime_start = row
                    continue

                current_regime = row['regime']
                if current_regime == last_regime:
                    continue
                else:
                    regime_end = row
                    data.append(
                        go.Scatter(
                            x=[regime_start['time'],regime_end['time']],
                            y=[30000, 30000],
                            line = dict(
                                color = (r_colors[regime_start['regime']])
                            ),
                            showlegend=False,
                            name = 'Regime {}'.format(regime_start['regime']),
                            fill='tozeroy'))

                    regime_start = regime_end


    if add_volume:
        volume = go.Bar(
            x = df['time'],	
            y = df['volume'], 
            xaxis="x", 
            yaxis="y2", 
            width = 400000,
            name = "Volume")

        data.append(volume)

    if add_candles:
        candle = go.Candlestick(
            x = df['time'],
            open = df['open'],
            close = df['close'],
            high = df['high'],
            low = df['low'],
            name = "Candlesticks")
        data.append(candle)
    else:
        price = go.Scatter( 
            x = df['time'],
            y = df['close'], 
            name = 'Price',
            line = dict(color = 'black'))

        data.append(price)
    
    for ind in plot_indicators:
        if df.__contains__(ind['name']):
            if ind.get('showlegend', None) is None:
                ind['showlegend'] = True
            if ind.get('xvalue', None) is None:
                ind['xvalue'] = 'time'
            if ind.get('color', None) is None:
                ind['color'] = None # 'rgba(102, 207, 255, 50)'
            if ind.get('yaxis', None) is None:
                ind['yaxis'] = 'y'
            if ind.get('xaxis', None) is None:
                ind['xaxis'] = 'x'
            if ind.get('type', None) is None:
                ind['type'] = 'scatter'
            if ind.get('width', None) is None:
                ind['width'] = 300000
            if ind.get('fill', None) is None:
                ind['fill'] = None
            if ind.get('mode', None) is None:
                ind['mode'] = 'lines'

            if ind['type'] == 'bar':
                trace = go.Bar(
                    x = df[ind['xvalue']], 
                    y = df[ind['name']], 
                    name = ind['title'],
                    xaxis = ind['xaxis'], 
                    yaxis = ind['yaxis'], 
                    marker_color = ind['color'], 
                    showlegend = ind['showlegend'],
                    width = ind['width'],
                marker = dict(color = ind['color']))
            else:
                trace = go.Scatter( 
                    x = df[ind['xvalue']],
                    y = df[ind['name']], 
                    name = ind['title'],
                    mode = ind['mode'], 
                    xaxis = ind['xaxis'], 
                    yaxis = ind['yaxis'], 
                    fill = ind['fill'], 
                    showlegend = ind['showlegend'],
                    line = dict(color = ind['color']))

            data.append(trace)

    if trend_points:
        mins = go.Scatter( x = df['time'], 	y = df['min'], name = "Min Points",
            line = dict(color = ('rgba(255, 100, 100, 255)')),
            mode = "markers",)
        data.append(mins)

        maxs = go.Scatter( x = df['time'], y = df['max'], name = "Max Points",
                line = dict(color = ('rgba(100, 255, 100, 255)')),
                mode = "markers",)
        data.append(maxs)

    if signals:
        for signal in signals:
            size_multiplier = [10 for z in signal["points"]]
            if len(signal["points"]) > 1:
                if len(signal["points"][0]) > 2:
                    size_multiplier = [z[2] for z in signal["points"]]

            scat = go.Scatter(
                x = [item[0] for item in signal['points']],
                y = [item[1] for item in signal['points']],
                name = signal['name'],
                mode = "markers",
                marker_size = size_multiplier 
            )
            data.append(scat)
    
    return data

def PlotData(df,
    add_candles:bool=True,
    add_volume:bool=True,
    buy_signals:bool or list=False,
    sell_signals:bool or list=False,
    signals:bool or list=False,
    trend_points=False,
    plot_indicators=[],
    plot_shapes=False,
    regimes_number=None,
    trends=False,
    save_plot=False,
    show_plot=False,
    plot_title:str="Unnamed"):
    '''
    Creates a plotly plot based on the options provided - which can be displayed
    in a front-end or saved as a standalone webpage.

    Params
    --
        buy signals: bool or list
            if not False it adds to the plot some points representing buy signals

        sell signals: bool or list
            if list, it adds to the plot some points representing sell signals

    '''

    data = GetPlotData(
        df,
        add_candles=add_candles,
        add_volume=add_volume,
        buy_signals=buy_signals,
        sell_signals=sell_signals,
        signals=signals,
        regimes_number=regimes_number,
        trend_points=trend_points,
        plot_indicators=plot_indicators,
        trends=trends)

    layout = go.Layout(
        # autosize=True,
        margin=dict(l=0, r=0, b=0, t=0),
        hovermode="closest",
        plot_bgcolor="#FFF",
        paper_bgcolor="#FFF",
        legend=dict(font=dict(size=8), orientation="h", x=0, y=0),
        # colorway=['#5E0DAC', '#FF4F00', '#375CB1', '#FF7400', '#FFF400', '#FF0056'],
        xaxis={
            'rangeslider': {
                'visible': False,
                'yaxis': {
                    'rangemode':'match'
                }
            }, 
            "showticklabels":False,
            'type': 'date'},
    )
    
    y2, y3 = False, False
    if add_volume:
        y2 = True
    for ind in plot_indicators:
        if ind.__contains__('yaxis'):
            if ind['yaxis'] == 'y2':
                y2 = True
            if ind['yaxis'] == 'y3':
                y3 = True

    if y2 and y3:
        layout.update(
            yaxis={
                "domain": [0.3, 1],
                # "title": "Price", 
                "fixedrange":False,
                "ticks": '',
                "showticklabels":False,
            })
        layout.update(
            yaxis2=dict(
                domain = [0.15, 0.29],
                side = 'right',
                showticklabels = False,
                # title  = "Volume"
            ))
        layout.update(
            yaxis3=dict(
                domain = [0, 0.15],
                showticklabels = False,
                # title=""
            ))
    elif y2:
        layout.update(
            yaxis={
                "domain": [0.25, 1],
                # "title": "Price", 
                "fixedrange":False,
                "ticks": '',
                "showticklabels":False},
        )
        layout.update(
            yaxis2=dict(
                domain = [0, 0.24],
                side = 'right',
                showticklabels = False,
                # title  = "Volume"
                ))
    elif y3:
        layout.update(
            yaxis={
                "domain": [0.25, 1],
                # "title": "Price", 
                "fixedrange":False,
                "ticks": '',
                "showticklabels":False},
        )
        layout.update(
            yaxis3=dict(
                domain = [0, 0.24],
                showticklabels = False,
                # ticks="",
                # title=""
                ))
    else:
        layout.update(
            yaxis={
                "domain": [0.1, 1],
                # "title": "Price", 
                "fixedrange":False,
                "ticks": '',
                "showticklabels":False},
        )

    if trends:
        layout.update(shapes=layout['shapes'].__add__(tuple(trends)))
    if plot_shapes:
        layout.update(shapes=layout['shapes'].__add__(tuple(plot_shapes)))

    # style and display
    fig = go.Figure(data = data, layout = layout)

    if save_plot or show_plot:
        file_path = os.path.abspath('graphs')
        if not os.path.exists(file_path):
            os.makedirs(file_path)
        plot(fig, filename=os.path.join(file_path, plot_title+'.html'), auto_open=show_plot)

    return fig


import math
import sqlalchemy as db
from datetime import datetime
from sqlalchemy import create_engine
from sqlalchemy.orm import backref, relationship, scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from pyjuque.Engine.Models.Utils import SqliteDecimal

Base = declarative_base()

def getScopedSession(path='sqlite:///'):
    some_engine = create_engine(path, echo=False)
    Base.metadata.create_all(some_engine)
    session_factory = sessionmaker(bind=some_engine)
    Session = scoped_session(session_factory)
    return Session

def getSession(path='sqlite:///', default_class=Base):
    some_engine = create_engine(path, echo=False)
    default_class.metadata.create_all(some_engine)
    Session = sessionmaker(bind=some_engine)
    session = Session()
    return session


class OrderModel(Base):
    __tablename__ = 'order'

    id = db.Column(db.String(32), primary_key=True)
    position_id = db.Column(db.Integer, index=True)
    bot_id = db.Column(db.Integer)
    symbol = db.Column(db.String(13), index=True)
    timestamp = db.Column(db.DateTime, index=True, default=datetime.now)
    price = db.Column(SqliteDecimal(13))
    take_profit_price = db.Column(SqliteDecimal(13), default=None)
    entry_price = db.Column(SqliteDecimal(13), default=None)
    stop_price = db.Column(SqliteDecimal(13), default=None)
    original_quantity = db.Column(SqliteDecimal(13))
    executed_quantity = db.Column(SqliteDecimal(13))
    status = db.Column(db.String(30), index=True)
    side = db.Column(db.String(30), index=True)
    is_entry = db.Column(db.Boolean, default=True)
    is_closed = db.Column(db.Boolean, default=False)
    matched_order_id = db.Column(db.Integer, db.ForeignKey('order.id'), default=None)
    is_test = db.Column(db.Boolean)
    order_type = db.Column(db.String(30), index = True)
    last_checked_time = db.Column(db.Integer)


class PairModel(Base):
    __tablename__ = 'pair'

    id = db.Column(db.Integer, primary_key=True)
    bot_id = db.Column(db.Integer, index=True)
    symbol = db.Column(db.String(13), index=True)
    active = db.Column(db.Boolean, default=True)
    current_order_id = db.Column(db.Integer, db.ForeignKey('order.id'))
    profit_loss = db.Column(SqliteDecimal(13), default=1)


class BaseBotModel(Base):

    __abstract__ = True
    __tablename__ = 'base_bot'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(30))  
    starting_balance = db.Column(SqliteDecimal(13))
    current_balance = db.Column(SqliteDecimal(13))
    profit_loss = db.Column(db.Float, default=100)
    test_run = db.Column(db.Boolean, default=False)

    def getOrders(self, session):
        return session.query(OrderModel).filter_by(bot_id=self.id).all()

    def getOpenOrders(self, session):
        return session.query(OrderModel).filter_by(bot_id=self.id, is_closed=False).all()    


class GridBotModel(BaseBotModel):
    __tablename__ = 'grid_bot'
   
    exchange = db.Column(db.String(30))
    symbol = db.Column(db.String(30))
    trade_amount = db.Column(SqliteDecimal(13))
    trade_step = db.Column(SqliteDecimal(13))


class TABotModel(BaseBotModel):
    __tablename__ = 'ta_bot'

    is_running = db.Column(db.Boolean, default=False)
    quote_asset = db.Column(db.String(10), index=True)
    entry_settings_id = db.Column(db.Integer, db.ForeignKey('entry_settings.id'))
    exit_settings_id = db.Column(db.Integer, db.ForeignKey('exit_settings.id'))

    def getPairs(self, session):
        return session.query(PairModel).filter_by(bot_id=self.id).all()

    def getActivePairs(self, session):
        return session.query(PairModel).filter_by(bot_id=self.id, active=True).all()

    def getPair(self, session, symbol=None):
        if symbol != None:
            return session.query(PairModel).filter_by(bot_id=self.id, symbol=symbol).first()
        return session.query(PairModel).filter_by(bot_id=self.id).first()

    def getFirstBuyOrder(self, session, position_id):
        return session.query(PairModel).filter_by(bot_id=self.id, position_id=position_id, side='BUY').first()
    

class EntrySettingsModel(Base):
    """
    Model for entry settings
    """

    __tablename__ = 'entry_settings'
    id = db.Column(db.Integer, primary_key=True)                  	# Unique ID
    name = db.Column(db.String(30))                                 # Name (for UI)
    bots = relationship('TABotModel', backref=backref('entry_settings'))
    open_buy_order_time_out = db.Column(db.Integer, default=math.inf)
    initial_entry_allocation = db.Column(db.Integer, default=None)	# What % of funds allocated to the bot will go to an initial entry
    subsequent_entries = db.Column(db.Integer, default=0)         	# Are there subsequent entries
    subsequent_entry_allocation = db.Column(db.Float, default=1)  	# What % of the initial quantity will we buy on a 
                                    # subsequent entry (1 = 100%, 0.5 = 50%)
    subsequent_entry_distance = db.Column(db.Float, default=None) 	# Distance between subsequent entries in %
    signal_distance = db.Column(db.Float, default=0)         		# Distance from signal for initial entry 
                                        # 0 means enter on signal, 1 means enter 1% away 
                                        # - in the direction opposite that of the trade - 
                                        # from the signal price


class ExitSettingsModel(Base):
    """ Exit Settings of a Bot """
    __tablename__ = 'exit_settings'
    id = db.Column(db.Integer, primary_key=True)                  	# Unique ID
    name = db.Column(db.String(30))                               	# Name for UI
    bots = relationship('TABotModel', backref=backref('exit_settings'))   # Name (for UI)
    profit_target = db.Column(db.Float)                           	# Exit when price is at value % profit from entry 
    stop_loss_value = db.Column(db.Float, default=None)             	# Whether to have stop loss or not (and what %)
    is_trailing_stop_loss = db.Column(db.Boolean, default=False)    	# Whether to have trailing stop loss or not (what %)
    stop_loss_active_after = db.Column(db.Float, default=None)    	# If we have trailing stop loss, whether to activate immediately, or after a value % increase in profit
    exit_on_signal = db.Column(db.Boolean, default=False)       



import sqlalchemy as db
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import backref, sessionmaker, scoped_session
from pyjuque.Engine.Models.Utils import SqliteDecimal

Base = declarative_base()

def getScopedSession(path='sqlite:///'):
    some_engine = create_engine(path, echo=False)
    Base.metadata.create_all(some_engine)
    session_factory = sessionmaker(bind=some_engine)
    Session = scoped_session(session_factory)
    return Session

def getSession(path='sqlite:///', default_class=Base):
    some_engine = create_engine(path, echo=False)
    default_class.metadata.create_all(some_engine)
    Session = sessionmaker(bind=some_engine)
    session = Session()
    return session


class CandlestickModel(Base):
    __tablename__ = 'candlestick'
    symbol = db.Column(db.String(13), index=True, primary_key=True)
    timeframe = db.Column(db.String(5), index=True, primary_key=True)
    timestamp = db.Column(db.String, index=True, primary_key=True)
    datetime = db.Column(db.DateTime)
    open = db.Column(db.String)
    high = db.Column(db.String)
    low = db.Column(db.String)
    close = db.Column(db.String)
    volume = db.Column(db.String)


import sys
from os import getenv
from yaspin import yaspin
from pprint import pprint
from decimal import Decimal
from pyjuque.Exchanges.CcxtExchange import CcxtExchange
from pyjuque.Engine.Models import GridBotModel, getSession
from pyjuque.Engine.OrderManager import *

import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class GridBotController:

    def __init__(self, name = None):
        self.status_printer = None
        self.screen = False
        if name == None:
            pass
            # print('GridBot is nonexistent, ' 
            #     'please call self.create() '
            #     'with the required parameters')
        else:
            session = getSession('sqlite:///{}.db'.format(name))
            bot_model = session.query(GridBotModel).first()
            if bot_model == None:
                pass
                # print('GridBot is nonexistent, '
                #     'please call self.create() '
                #     'with the required parameters')
            else:
                # print('GridBot found!')
                self.bot_model = bot_model
                self.name = name
                self.session = session
                self.symbol = bot_model.symbol
                self.total_amount = bot_model.starting_balance
                self.trade_amount = bot_model.trade_amount
                self.trade_step = bot_model.trade_step
                self.test_mode = bot_model.test_run
                self.exchange_name = bot_model.exchange.upper()
                self.exchange = CcxtExchange(bot_model.exchange.lower(), {
                    'apiKey': getenv('{}_API_KEY'.format(self.exchange_name)), 
                    'secret': getenv('{}_API_SECRET'.format(self.exchange_name)),
                    'password': getenv('{}_PASSWORD'.format(self.exchange_name)),
                    'timeout': 30000,
                    # 'verbose': True,
                    'enableRateLimit': True,
                })


    def create(self, exchange, symbol, total_amount, trade_amount, trade_step, total_trades, test_mode=False):
        self.exchange = exchange
        self.exchange_name = self.exchange.exchange_id.upper()
        self.symbol = symbol
        self.total_amount = total_amount
        self.trade_amount = trade_amount
        self.trade_step = trade_step
        self.total_trades = total_trades
        self.test_mode = test_mode
        self.name = 'GridBot_{}_{}_{}_{}_{}'.format(
            self.exchange_name, 
            symbol.replace('/', ''), 
            str(int(total_amount)), 
            str(int(trade_amount*total_amount)), 
            str(trade_step*100).replace('.', 'point'))
        if test_mode:
            self.name = self.name+'_test'
            raise NotImplementedError('Test Mode not implemented for GridBot')
        else:
            self.name = self.name+'_live'
        # print('Bot name is {}'.format(self.name))
        self.session = getSession('sqlite:///{}.db'.format(self.name))

        bot_model = self.session.query(GridBotModel).first()
        if bot_model == None:
            self._initializeDatabase()
        else:
            self.bot_model = bot_model


    def _initializeDatabase(self):
        """ Function that initializes the database
        by creating a bot with two pairs. """
        self.bot_model = GridBotModel(
            name = self.name,
            symbol = self.symbol,
            exchange = self.exchange.exchange_id,
            starting_balance = self.total_amount,
            current_balance = self.total_amount,
            trade_amount = self.trade_amount,
            trade_step = self.trade_step,
            test_run = self.test_mode)
        self.session.add(self.bot_model)
        self.session.commit()


    def executeBot(self):
        ticker = self.exchange.ccxt.fetchTicker(self.symbol)
        last_price = Decimal(ticker['last'])
        open_orders = self.bot_model.getOpenOrders(self.session)
        if len(open_orders) == 0:
            # Place Initial Buy Orders
            self.status_printer.text = 'Palcing Initial Orders'
            self.placeInitialOrders(last_price)
            open_orders = self.bot_model.getOpenOrders(self.session)
        for order in open_orders:
            self.status_printer.text = 'Checking Placed Orders'
            self.updateOpenOrder(order, last_price)
        self.updateLastOrder(last_price)
        self.session.commit()


    def placeInitialOrders(self, last_price):
        for i in range(self.total_trades):
            buy_price = last_price - last_price * Decimal(i+1) * Decimal(self.trade_step)
            order = self.placeOrder(
                symbol=self.symbol, side='buy', price=buy_price,  
                quantity=self.trade_amount, order_type='limit', 
                is_entry=True)
            order.position_id = i
        self.session.commit()


    def updateOpenOrder(self, order, last_price):
        if not self.test_mode: 
            # try:
            exchange_order_info = self.exchange.getOrder(order.symbol, order.id, is_custom_id=True)
            # except Exception:
            #     self.log('Error getting data from the exchange for updating open order on {}.'.format(self.symbol))
            #     self.log(sys.exc_info())
            #     return
        else:
            raise NotImplementedError('Test Mode not implemented for GridBot')
        
        # print('Order Info:')
        # pprint(exchange_order_info)
        order.side = exchange_order_info['side']
        order.status = exchange_order_info['status']
        order.executed_quantity = exchange_order_info['filled']
        if exchange_order_info['fee'] != None:
            order.executed_quantity = exchange_order_info['filled'] + exchange_order_info['fee']['cost']
        if (order.side == 'buy'):
            # buy order was filled, place exit order.
            if (order.status == 'closed'):
                self.screen.clear()
                self.screen.refresh()
                # pprint(exchange_order_info)
                self.log('Buy order at {} filled, place exit.'.format(order.price))
                self.placeExitOrder(order)
                self.placeFarthestEntryOrder(last_price)
            elif (order.status in ['canceled', 'expired', 'rejected']):
                if order.executed_quantity > 0:
                    self.log('Buy order at {} canceled, but partially filled, place exit.'.format(order.price))
                    self.placeExitOrder(order)
                    self.placeFarthestEntryOrder(last_price)
                else:
                    order.is_closed = True
        # sell order 
        if (order.side == 'sell'):
            # sell order was filled
            if (order.status == 'closed'):
                self.screen.clear()
                self.screen.refresh()
                # pprint(exchange_order_info)
                self.log('Sell order at {} filled, cancel farthest buy order and replace with new buy order.'.format(order.price))
                self.cancelFarthestEntryOrder(last_price)
                self.placeEntryOrder(order)
                order.is_closed = True
            # sell order was rejected by engine of exchange
            # if (order.status in ['rejected', 'expired', 'canceled']):
            #     original_buy_order = self.reviveOriginalBuyOrder(order)
            #     self.placeExitOrder(original_buy_order)


    def placeEntryOrder(self, exit_order):
        # entry_order = self.session.query(Order).filter_by(matched_order_id=exit_order.id)
        buy_price = exit_order.price * Decimal(1 - self.trade_step)
        self.placeOrder(
            symbol = self.symbol,
            price = buy_price, 
            quantity = exit_order.executed_quantity, 
            side = 'buy', 
            order_type = 'limit', 
            is_entry=True)


    def placeExitOrder(self, entry_order):
        sell_price = entry_order.price * Decimal(1 + 2 * self.trade_step)
        self.placeOrder(
            symbol = self.symbol,
            entry_order = entry_order, 
            price = sell_price, 
            quantity = entry_order.executed_quantity, 
            side = 'sell', 
            order_type = 'limit')


    def updateLastOrder(self, last_price):
        open_orders = [order for order in self.bot_model.getOpenOrders(self.session) if order.side == 'buy']
        if len(open_orders) == 0:
            return
        farthest_order = None
        closest_order = None
        max_difference = 0
        min_difference = 10000
        for order in open_orders:
            difference = last_price - order.price
            if difference > max_difference:
                max_difference = difference
                farthest_order = order
            if difference < min_difference:
                min_difference = difference
                closest_order = order
        if min_difference > last_price * Decimal(2 * self.trade_step):
            order_result = self.exchange.cancelOrder(
                farthest_order.symbol, farthest_order.id, is_custom_id=True)
            farthest_order.status = 'canceled'
            farthest_order.is_closed = True
            buy_price = closest_order.price * Decimal(1 + self.trade_step)
            self.placeOrder(
                symbol = self.symbol,
                price = buy_price, 
                quantity = farthest_order.original_quantity, 
                side = 'buy', 
                order_type = 'limit',
                is_entry=True)


    def cancelFarthestEntryOrder(self, last_price):
        open_orders = [order for order in self.bot_model.getOpenOrders(self.session) if order.side == 'buy']
        if len(open_orders) == 0:
            return
        farthest_order = None
        max_difference = 0
        for order in open_orders:
            difference = last_price - order.price
            if difference > max_difference:
                max_difference = difference
                farthest_order = order
        order_result = self.exchange.cancelOrder(farthest_order.symbol, farthest_order.id, is_custom_id=True)
        farthest_order.status = 'canceled'
        farthest_order.is_closed = True


    def placeFarthestEntryOrder(self, last_price):
        open_orders = [order for order in self.bot_model.getOpenOrders(self.session) if order.side == 'buy']
        if len(open_orders) == 0:
            buy_price = last_price * Decimal(1 - self.trade_step)
            self.placeOrder(
                symbol = self.symbol,
                price = buy_price, 
                quantity = self.trade_amount, 
                side = 'buy', 
                order_type = 'limit',
                is_entry=True)
            return
        farthest_order = None
        max_difference = 0
        for order in open_orders:
            difference = last_price - order.price
            if difference > max_difference:
                max_difference = difference
                farthest_order = order
        buy_price = farthest_order.price * Decimal(1 - self.trade_step)
        self.placeOrder(
            symbol = self.symbol,
            price = buy_price, 
            quantity = self.trade_amount, 
            side = 'buy', 
            order_type = 'limit',
            is_entry=True)


    def placeOrder(self, entry_order=None, **order_params):
        """ Create Order model and place order to exchange. """
        order_params['bot_id'] = self.bot_model.id
        new_order = placeNewOrder(
            exchange = self.exchange, 
            symbol = self.symbol, 
            order = entry_order, 
            test_mode = self.test_mode,
            order_params = order_params)
        if new_order != None:
            self.syncModels(entry_order, new_order)
            self.session.add(new_order)
        return new_order


    def syncModels(self, entry_order, new_order):
        """ Sync orders to new status """
        if entry_order is not None:
            new_order.matched_order_id = entry_order.id
            entry_order.is_closed = True
            entry_order.matched_order_id = new_order.id
        else:
            self.bot_model.current_balance = self.bot_model.current_balance \
                - Decimal(self.trade_amount) * Decimal(self.bot_model.starting_balance)


    def log(self, message, should_print=True):
        # if self.screen:
        #     self.screen.clear()
        #     self.screen.refresh()
        if self.status_printer != None:
            self.status_printer.stop()
            if should_print:
                logger.info(message)
            else:
                self.status_printer.text = message
            self.status_printer.start()
        elif should_print:
            logger.info(message)



class BaseExchange:
    """
        Is the base class from which all Exchanges should inherit.
        Exchanges should implement all the functions outlined here (at least).

        As we add more exchanges to the bot & see different API implementations, 
        we will also update this class - so that it has the most elastic & modular 
        form, in case we observe that more functions are needed or that some need 
        to be updated.
    """

    def _get(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Calls a get request to this exchange """
        raise NotImplementedError

    def _post(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Calls a post request to this exchange """
        raise NotImplementedError

    def _delete(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Calls a delete request to this exchange """
        raise NotImplementedError

    def _signRequest(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Signs a request with the API & SECRET keys """
        raise NotImplementedError

    def addCredentials(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Adds API & SECRET keys into the object's memory """
        raise NotImplementedError

    def getAccountData(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Gets data for Account connected to API & SECRET Keys given """
        raise NotImplementedError
    
    def getTradingSymbols(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Gets All symbols which are tradable (currently) """
        raise NotImplementedError

    def getOrderBook(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Gets Order Book data for symbol """
        raise NotImplementedError

    def getOHLCV(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Gets Candlestick Book data for symbol on interval (IE 1 minute) """
        raise NotImplementedError

    def placeOrder(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Places order on exchange given a dictionary of parameters """
        raise NotImplementedError

    def placeMarketOrder(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Places side (buy/sell) market order for amount of symbol  """
        raise NotImplementedError

    def placeLimitOrder(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Places side (buy/sell) limit order for amount of symbol at price """
        raise NotImplementedError

    def cancelOrder(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Cancels order given order id """
        raise NotImplementedError

    def getOrder(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Gets order info given order id """
        raise NotImplementedError

    def isValidResponse(self, *args, **kwargs):
        """ (NOT IMPLEMENTED) 
        Checks whether response received from exchange is valid

        Returns 
        --
            True if valid, False otherwise
        """
        raise NotImplementedError
    


from abc import ABC, abstractmethod

class StrategyTemplate(ABC):

    minimum_period = 100
    indicators = []
    df = None

    @abstractmethod
    def setUp(self, df):
        """ Computes the indicators for the given dataframe 
        (usually gets called by the bot controller when recent candlestick 
        data is received from the exchange)"""
        pass


    @abstractmethod
    def checkLongSignal(self, i):
        """ Checks whether we have a long signal """
        pass


    @abstractmethod
    def checkShortSignal(self, i):
        """ Checks whether we have a short signal """
        pass




import sqlalchemy.types as types
from decimal import Decimal


class SqliteDecimal(types.TypeDecorator):
  # This TypeDecorator use Sqlalchemy Integer as impl. It converts Decimals
  # from Python to Integers which is later stored in Sqlite database.
  impl = types.BigInteger

  def __init__(self, scale):
    # It takes a 'scale' parameter, which specifies the number of digits
    # to the right of the decimal point of the number in the column.
    types.TypeDecorator.__init__(self)
    self.scale = scale
    self.multiplier_int = 10 ** self.scale

  def process_bind_param(self, value, dialect):
    # e.g. value = Column(SqliteDecimal(2)) means a value such as
    # Decimal('12.34') will be converted to 1234 in Sqlite
    if value is not None:
      value = int(Decimal(value) * self.multiplier_int)
    return value

  def process_result_value(self, value, dialect):
    # e.g. Integer 1234 in Sqlite will be converted to SqliteDecimal('12.34'),
    # when query takes place.
    if value is not None:
      value = Decimal(value) / self.multiplier_int
    return value




import requests
import json
import hashlib
import hmac
import time
import pandas
from decimal import Context, Decimal, getcontext
from traceback import print_exc
import math

import os
import sys
curr_path = os.path.abspath(__file__)
root_path = os.path.abspath(os.path.join(curr_path, os.path.pardir, os.path.pardir))
sys.path.append(root_path)
from os.path import join, dirname

from dotenv import load_dotenv, find_dotenv
load_dotenv(find_dotenv())
from datetime import datetime

from pyjuque.Exchanges.Base.Exceptions import \
    InvalidCredentialsException, \
    InternalExchangeException, \
    ExchangeConnectionException

from pprint import pprint
from enum import Enum

class Binance():
    """ Wrapper around the Binance REST API """

    ORDER_STATUS_NEW = 'NEW'
    ORDER_STATUS_PARTIALLY_FILLED = 'PARTIALLY_FILLED'
    ORDER_STATUS_FILLED = 'FILLED'
    ORDER_STATUS_CANCELED = 'CANCELED'
    ORDER_STATUS_PENDING_CANCEL = 'PENDING_CANCEL' # NOT USED BY BINANCE ATM
    ORDER_STATUS_REJECTED = 'REJECTED'
    ORDER_STATUS_EXPIRED = 'EXPIRED'

    ORDER_SIDE_BUY = 'BUY'
    ORDER_SIDE_SELL = 'SELL'

    ORDER_TYPE_LIMIT = 'LIMIT'
    ORDER_TYPE_MARKET = 'MARKET'
    ORDER_TYPE_STOP_LOSS = 'STOP_LOSS'
    ORDER_TYPE_STOP_LOSS_LIMIT = 'STOP_LOSS_LIMIT'
    ORDER_TYPE_TAKE_PROFIT = 'TAKE_PROFIT'
    ORDER_TYPE_TAKE_PROFIT_LIMIT = 'TAKE_PROFIT_LIMIT'
    ORDER_TYPE_LIMIT_MAKER = 'LIMIT_MAKER'

    KLINE_INTERVALS = ['1m', '3m', '5m', '15m', '30m', '1h', '2h', '4h', '6h', '8h', '12h', '1d', '3d', '1w', '1M']

    BASE_URL = 'https://api.binance.com'
    
    ENDPOINTS = {
        "order": '/api/v3/order',
        "testOrder": '/api/v3/order/test',
        "allOrders": '/api/v3/allOrders',
        "klines": '/api/v3/klines',
        "exchangeInfo": '/api/v3/exchangeInfo',
        "averagePrice" : '/api/v3/avgPrice',
        "orderBook" : '/api/v3/depth',
        "account" : '/api/v3/account',
        "ticker": '/api/v3/ticker/bookTicker',
        "tickprice": '/api/v3/ticker/24hr'
    }

    SYMBOL_DATAS = dict()
    TICKER_DATA = dict()
    TICKER_UPDATE_TIME = None

    def __init__(self, filename=None, api_key=None, secret_key=None, get_credentials_from_env=False):
        
        self.api_keys = None
        self.has_credentials = False

        self.updateSymbolsData()
        self.updateTickerData()

        if get_credentials_from_env:
            env_api_key = os.getenv('BINANCE_API_KEY')
            env_secret_key = os.getenv('BINANCE_API_SECRET')
            if env_api_key is not None and env_secret_key is not None:
                self.addCredentials(env_api_key, env_secret_key)
        else:
            if filename is not None:
                file = open(filename, "r")
                if file.mode == 'r':
                    contents = file.read().split('\n')
                    self.addCredentials(api_key = contents[0], secret_key=contents[1])
                else:
                    raise Exception("Can't read", filename, "make sure the file is readable!")
        
        # Adding credentials FROM API & SECRET KEYS passed through init
        if self.has_credentials is False:
            if api_key is not None and secret_key is not None:
                self.addCredentials(api_key, secret_key)

        self.headers = dict()
        # Adding required Headers if credentials exist
        if self.has_credentials:
            self.headers["X-MBX-APIKEY"] = self.api_keys['api_key']

    def _get(self, url, params=None, headers=None):
        """ Implements a get request for this exchange """
        try:
            response = requests.get(url, params=params, headers=headers)
            payload = json.loads(response.text)
            if type(payload) == type([]):
                data = dict(payload=payload, url=url)
            else:
                data = payload
                data['url'] = url

        except requests.exceptions.ConnectionError:
            raise ExchangeConnectionException()
        except Exception as e:
            print("Exception occured when trying to GET from "+url)
            print_exc()
            data = {'code': '-1', 'url':url, 'msg': e}
            # raise ExchangeConnectionException()
        return data

    def _post(self, url, params=None, headers=None):
        """ Implements a post request for this exchange """
        try: 
            response = requests.post(url, params=params, headers=headers)
            data = json.loads(response.text)
            data['url'] = url
        except Exception as e:
            print("Exception occured when trying to POST to "+url); print(e); print("Params"); print(params)
            print_exc()
            data = {'code': '-1', 'url':url, 'msg': e}
        return data

    def _delete(self, url, params=None, headers=None):
        """ Implements a delete request for this exchange """
        try: 
            response = requests.delete(url, params=params, headers=headers)
            data = json.loads(response.text)
            data['url'] = url
        except Exception as e:
            print("Exception occured when trying to DELETE from "+url); print(e); print("Params"); print(params)
            print_exc()
            data = {'code': '-1', 'url':url, 'msg': e, 'params':params}
        return data

    def _signRequest(self, params):
        """ Signs a request with the API & SECRET keys """
        query_string = '&'.join(["{}={}".format(d, params[d]) for d in params])
        signature = hmac.new(self.api_keys['secret_key'].encode('utf-8'), query_string.encode('utf-8'), hashlib.sha256)
        params['signature'] = signature.hexdigest()

    def _getSymbolsData(self):
        """ Gets All symbols' data from Binance """
        url = Binance.BASE_URL + Binance.ENDPOINTS["exchangeInfo"]
        data = self._get(url)
        if not Binance.isValidResponse(data):
            return []

        symbols_list = []
        for pair in data['symbols']:
            symbols_list.append(pair)

        return symbols_list

    def updateSymbolsData(self):
        symbols = self._getSymbolsData()
        for symb in symbols:
            Binance.SYMBOL_DATAS[symb['symbol']] = symb

    def _getTickerData(self):
        """ Gets All symbols' data from Binance """
        url = Binance.BASE_URL + Binance.ENDPOINTS["ticker"]
        data = self._get(url)
        if not Binance.isValidResponse(data):
            return []
        
        symbols_list = []
        for pair in data["payload"]:
            symbols_list.append(pair)

        return symbols_list

    def updateTickerData(self):
        symbols = self._getTickerData()
        self.TICKER_UPDATE_TIME = time.time()
        for symb in symbols:
            Binance.TICKER_DATA[symb["symbol"]] = symb

    def addCredentials(self, api_key, secret_key):
        """ Adds API & SECRET keys into the object's memory """
        new_keys = dict(api_key=api_key, secret_key=secret_key)
        self.api_keys = new_keys
        self.has_credentials = True

    def getCurrentTickPrice(self, symbol):
        url = Binance.BASE_URL + Binance.ENDPOINTS['tickprice']
        params = {'symbol': symbol.upper()}
        tick_info = self._get(url, params, self.headers)
        return tick_info

    def getAccountData(self):
        """ Gets data for Account connected to API & SECRET Keys given """	
        
        url = Binance.BASE_URL + Binance.ENDPOINTS["account"]
        params = {
        'recvWindow': 5000,
        'timestamp': int(round(time.time()*1000))
        }
        self._signRequest(params)
        return self._get(url, params, self.headers)

    def getTradingSymbols(self, quote_assets:list=None):
        """ Gets All symbols which are tradable (currently) 
            Params
            --
                `quote_assets`
                    a list of all the quote assets we are interested in;
                    if empty or None, we will return all trading pairs, 
                    indifferent of quote asset
            Returns
            --
                a list of symbols that are currently trading on Binance 
        """
        url = Binance.BASE_URL + Binance.ENDPOINTS["exchangeInfo"]
        data = self._get(url)
        if not Binance.isValidResponse(data):
            raise ExchangeConnectionException()

        symbols_list = []
        if quote_assets != None and len(quote_assets) > 0:
            for pair in data['symbols']:
                if pair['status'] == 'TRADING':
                    if pair['quoteAsset'] in quote_assets:
                        symbols_list.append(pair['symbol'])
        else:
            for pair in data['symbols']:
                if pair['status'] == 'TRADING':
                    symbols_list.append(pair['symbol'])

        return symbols_list

    def _getPriceInBTCDirectly(self, asset:str):
        """ Returns the price of an `asset` that is traded with BTC as quote or base asset. """
        price_in_btc = None
        getcontext().prec = 28
        if asset == 'BTC':
            price_in_btc = Decimal(1)

        elif self.TICKER_DATA.__contains__(asset+"BTC"):
            ticker_data = self.TICKER_DATA[asset+"BTC"]
            ticker_price = Decimal(0.5) * \
                (Decimal(ticker_data['askPrice']) + Decimal(ticker_data['bidPrice']))
            price_in_btc = ticker_price
            if ticker_price == 0:
                price_in_btc = None
            else:
                price_in_btc = ticker_price

        elif self.TICKER_DATA.__contains__("BTC"+asset):
            ticker_data = self.TICKER_DATA["BTC"+asset]
            ticker_price = Decimal(0.5) * \
                (Decimal(ticker_data['askPrice']) + Decimal(ticker_data['bidPrice']))
            if ticker_price == 0:
                price_in_btc = None
            else:
                price_in_btc = Decimal(1) / ticker_price

        return price_in_btc

    def getPriceInBTC(self, asset:str):
        """ Returns the price of an 'asset', uses triangulation to determine BTC price if needed."""
        possible_pairs = dict()
        price_in_btc = None
        findBTCDirectly = False
        getcontext().prec = 28
        # loop through all pairs to see if asset is traded with BTC as quote or base asset
        for key, value in self.SYMBOL_DATAS.items():
            if asset in key:
                possible_pairs[key] = dict(base=value['baseAsset'], quote=value['quoteAsset'])
                if 'BTC' in key:
                    findBTCDirectly = True

        if findBTCDirectly:
            # asset is traded with BTC as quote or base asset, find price directly.
            price_in_btc = self._getPriceInBTCDirectly(asset)
        elif not findBTCDirectly:
            # asset is not traded with BTC as quote or base asset, use triangulation to find BTC price.
            for pair, quote_base in possible_pairs.items():

                ticker_data = self.TICKER_DATA[pair]
                ticker_price = Decimal(0.5) * \
                (Decimal(ticker_data['askPrice']) + Decimal(ticker_data['bidPrice']))

                if asset == quote_base['base']:
                    price_asset_in_new_asset = ticker_price
                    new_asset = quote_base['quote']
                if asset == quote_base['quote']:
                    price_asset_in_new_asset = Decimal(1) / ticker_price
                    new_asset = quote_base['base']

                btc_price_new_asset = self._getPriceInBTCDirectly(new_asset)

                if btc_price_new_asset is None:
                    # new_asset does not have a quote or base asset in BTC
                    continue
                if btc_price_new_asset is not None:
                    # new_asset does have a quote or base asset in BTC, we can now calculate BTC price!
                    if price_asset_in_new_asset == 0:
                        price_in_btc = None
                    else:
                        price_in_btc = price_asset_in_new_asset * btc_price_new_asset
                        break
        return price_in_btc
    
    def getOrderBook(self, symbol, limit=100):
        """ Gets Order Book data for symbol """

        url = Binance.BASE_URL + Binance.ENDPOINTS["orderBook"]\
            + "?&symbol=" + str(symbol) + "&limit=" + str(limit)
        
        data = self._get(url)

        if not Binance.isValidResponse(data):
            print(data)
            raise Exception(data['msg'])

        return data
    
    def getOrderBookAveragePrice(self, symbol, side, quantity, order_book=None):
        #TODO test it
        """ Calculates the average price we would pay / receive per unit of `symbol` 
        if we wanted to trade `quantity` of that `symbol`, based on its order book"""
        
        obl_index = 0	# index of current order_book limit
        order_book_limits = [100, 500, 1000, 5000]	# possible order_book limits
        if order_book is not None:
            order_book = self.getOrderBook(symbol=symbol, limit=order_book_limits[obl_index])

        # print("Initial order book looks like")
        # pprint(order_book)
        order_book_side = order_book['asks'] \
            if side == Binance.ORDER_SIDE_SELL else order_book['bids']

        quantity = Decimal(quantity)
        i, orders, price = 0, [], Decimal(0)
        accounted_for_quantity = Decimal(0)

        qtdif = Decimal(1)
        while accounted_for_quantity < quantity or qtdif > Decimal(0.0001):
            try:
                order = order_book_side[i]
            except IndexError:
                raise InternalExchangeException("There are not enough orders in the Order Book.")
            qty = min(Decimal(order[1]), quantity - accounted_for_quantity)
            price += Decimal(order[0]) * qty
            accounted_for_quantity += qty
            qtdif = abs(Decimal(1) - accounted_for_quantity / quantity)
            i += 1
            if i >= order_book_limits[obl_index] - 1:
                # print("Reached Order Book Limit of ", i, "Try to get some more")
                if obl_index < 3:
                    obl_index += 1
                    order_book = self.getOrderBook(symbol=symbol, limit=order_book_limits[obl_index])
                    # print("Got new order book with limit", order_book_limits[obl_index])
                    order_book_side = order_book['asks'] \
                        if side == Binance.ORDER_SIDE_SELL else order_book['bids']
                    # print(i, len(order_book_side))
                    # print(qtdif)
                else:
                    # print(symbol, side, quantity)
                    raise InternalExchangeException("There are not enough orders in the Order Book.")

        return price / quantity

    def _getSymbolKlinesExtra(self, symbol:str, interval:str, 
    limit:int=1000, end_time=False, cast_to:type=float):
        """ Gets candlestick data for one symbol if more than 1000
        candles are requested (because of the Binance rate limitation
        of only 1000 candles per request, this function cals getSymbolKlines
        as many times it needs to get the desired number of candles)
        Parameters
        --
            Same as getSymbolKlines
        Returns
        --
            Same as getSymbolKlines
        """

        # Basicall, we will be calling the GetSymbolKlines as many times as we need 
        # in order to get all the historical data required (based on the limit parameter)
        # and we'll be merging the results into one long dataframe.
        repeat_rounds = 0
        if limit > 1000:
            repeat_rounds = int(limit/1000)
        initial_limit = limit % 1000
        if initial_limit == 0:
            initial_limit = 1000
        # First, we get the last initial_limit candles, starting at end_time and going
        # backwards (or starting in the present moment, if end_time is False)
        df = self.getOHLCV(
            symbol, interval, limit=initial_limit, end_time=end_time, cast_to=cast_to)
        while repeat_rounds > 0:
            # Then, for every other 1000 candles, we get them, but starting at the beginning
            # of the previously received candles.
            df2 = self.getOHLCV(
                symbol, interval, limit=1000, end_time=df['time'][0], cast_to=cast_to)
            df = df2.append(df, ignore_index = True)
            repeat_rounds = repeat_rounds - 1
        
        return df

    def getOHLCV(self, symbol:str, interval:str, 
    limit:int=1000, end_time:any=False, start_time:any=False, cast_to:type=float):
        """
        Gets candlestick data for one symbol 
        
        Parameters
        --
            symbol str:
                The symbol for which to get the trading data
            interval str:
                The interval on which to get the trading data
                minutes: '1m' '3m' '5m' '15m' '30m' 
                hours: '1h' '2h' '4h' '6h' '8h' '12h'
                days: '1d' '3d' weeks: '1w' months: '1M'
            limit int: 
                The number of data points (candles) to return
            `end_time` bool/int:
                The time of the last candle (if False, uses current time)
            
        Returns
        --
            DataFrame containing OHLCV data.
        """

        if limit > 1000:
            return self._getSymbolKlinesExtra(
                symbol, interval, limit, end_time, cast_to=cast_to)
        
        params = '?&symbol='+symbol+'&interval='+interval+'&limit='+str(limit)
        if end_time:
            params = params + '&endTime=' + str(int(end_time))
        if start_time:
            params = params + '&startTime=' + str(int(start_time))

        url = Binance.BASE_URL + Binance.ENDPOINTS['klines'] + params

        # download data
        try:
            data = requests.get(url)
        except requests.exceptions.ConnectionError:
            raise ExchangeConnectionException()
        dictionary = json.loads(data.text)

        # put in dataframe and clean-up
        df = pandas.DataFrame.from_dict(dictionary)
        df = df.drop(range(6, 12), axis=1)

        # rename columns
        col_names = ['time', 'open', 'high', 'low', 'close', 'volume']
        df.columns = col_names

        # transform values from strings to floats
        for col in col_names:
            df[col] = df[col].astype(cast_to)

        df['date'] = pandas.to_datetime(df['time'] * 1000000, infer_datetime_format=True)

        return df

    def placeOrder(self, params, test:bool=False, verbose = True):
        """ Places order on exchange given a dictionary of parameters.
            Check this link for more info on the required parameters: 
            https://github.com/binance-exchange/binance-official-api-docs/blob/master/rest-api.md#new-order--trade
            Parameters
            --
                params dict:
                    A dictionary of parameters containing info about the order,
                    such as SIDE (buy/sell), TYPE (market/limit/oco), SYMBOL, etc.
                test bool:
                    Decides whether this will be a test order or not.
        """

        params['symbol'] = params['symbol'].upper()

        self._signRequest(params)
        
        if verbose:
            print("\n\nPlacing Order with params:")
            pprint(params)
            print("\n")
        url = Binance.BASE_URL + Binance.ENDPOINTS['order']
        if test: 
            url = Binance.BASE_URL + Binance.ENDPOINTS['testOrder']

        return self._post(url, params, self.headers)

    def placeMarketOrder(self, symbol:str, side:str, amount:str, quote_amount=None, 
    test:bool=False, round_up_amount=False, custom_id=False, verbose=False):
        """ Places side (buy/sell) market order for amount of symbol.	"""
        if verbose:
            print("Placing MarketOrder with params:")
        
        params = {
            'symbol': symbol.upper(),
            'side': side,
            'type': Binance.ORDER_TYPE_MARKET,
            'recvWindow': 5000,
            'timestamp': int(round(time.time()*1000))
        }

        if custom_id is not False:
            params['newClientOrderId'] = custom_id

        if amount is not None:
            params['quantity'] = format(self.toValidQuantity(symbol, 
                                                                        amount, round_up_amount), 'f')
        elif quote_amount is not None:
            params['quoteOrderQty'] = format(self.toValidQuantity(symbol, 
                                                                quote_amount, round_up_amount), 'f')
        if verbose:
            pprint(params)
            
        return self.placeOrder(params, test)

    def placeLimitOrder(self, symbol:str, price, side:str, amount, 
    quote_amount=None, test:bool=False, round_up_price=False, time_in_force=None,
    round_up_amount=False, custom_id=False, verbose=False):
        """ Places side (buy/sell) limit order for amount of symbol at price. """
        if verbose:
            print("Placing LimitOrder with params:")
        
        params = {
            'symbol': symbol.upper(),
            'side': side,
            'type': Binance.ORDER_TYPE_LIMIT,
            'recvWindow': 5000,
            'timeInForce': 'GTC',
            'price': format(self.toValidPrice(symbol, price, round_up_price), 'f'),
            'timestamp': int(round(time.time()*1000))
        }
                
        if custom_id is not False:
            params['newClientOrderId'] = custom_id
                
        if time_in_force in ['GTC', 'IOC', 'FOK']:
            params['timeInForce'] = time_in_force

        if amount is not None:
            params['quantity'] = format(self.toValidQuantity(symbol, 
                                                                        amount, round_up_amount), 'f')
        elif quote_amount is not None:
            params['quoteOrderQty'] = format(self.toValidQuantity(symbol, 
                                                                quote_amount, round_up_amount), 'f')
        if verbose:
            pprint(params)
        return self.placeOrder(params, test)

    def placeStopLossMarketOrder(self, symbol:str, price, side:str,  amount, 
    quote_amount=None, test:bool=False, round_up_price=False, round_up_amount=False, 
    custom_id=False, verbose=False):
        """ Places a STOP_LOSS market order for amount of symbol at price. """
        if verbose:
            print("Placing StopLossMarketOrder with params:")

        params = {
            'symbol': symbol.upper(),
            'side': side,
            'type': Binance.ORDER_TYPE_STOP_LOSS,
            'recvWindow': 5000,
            'stopPrice': format(self.toValidPrice(symbol, price, round_up_price), 'f'),
            'timestamp': int(round(time.time()*1000))
        }
                
        if custom_id is not False:
            params['newClientOrderId'] = custom_id
            
        if amount is not None:
            params['quantity'] = format(self.toValidQuantity(symbol, amount, round_up_amount), 'f')
        
        elif quote_amount is not None:
            params['quoteOrderQty'] = format(self.toValidQuantity(symbol, 
                                                                quote_amount, round_up_amount), 'f')

        if verbose:
            pprint(params)

        return self.placeOrder(params, test)

    def placeStopLossLimitOrder(self, symbol:str, price, stop_price, side:str, amount, quote_amount=None, 
    test:bool=False, round_up_price=False, round_up_amount=False, custom_id=False, verbose=False):
        """ Places a STOP_LOSS_LIMIT market order for amount of symbol at price. """
        if verbose:
            print("Placing StopLossLimitOrder with params:")

        params = {
            'symbol': symbol.upper(),
            'side': side,
            'type': Binance.ORDER_TYPE_STOP_LOSS_LIMIT,
            'recvWindow': 5000,
            'timeInForce': 'GTC',
            'price': format(self.toValidPrice(symbol, price, round_up_price), 'f'),
            'stopPrice': format(self.toValidPrice(symbol, stop_price, round_up_price), 'f'),
            'timestamp': int(round(time.time()*1000))
        }
                
        if custom_id is not False:
            params['newClientOrderId'] = custom_id
                
        if amount is not None:
            params['quantity'] = format(self.toValidQuantity(symbol, 
                                                                        amount, round_up_amount), 'f')
        elif quote_amount is not None:
            params['quoteOrderQty'] = format(self.toValidQuantity(symbol, 
                                                                quote_amount, round_up_amount), 'f')

        if verbose:
            pprint(params)

        return self.placeOrder(params, test)

    def placeTakeProfitMarketOrder(self, symbol:str, price, side:str, amount, quote_amount=None, 
    test:bool=False, round_up_price=False, round_up_amount=False, custom_id=False, verbose=False):
        """ Places a TAKE_PROFIT market order for amount of symbol at price. """
        if verbose:
            print("Placing TakeProfitMarketOrder with params:")
        
        params = {
            'symbol': symbol.upper(),
            'side': side,
            'type': Binance.ORDER_TYPE_TAKE_PROFIT,
            'recvWindow': 5000,
            'stopPrice': format(self.toValidPrice(symbol, price, round_up_price), 'f'),
            'timestamp': int(round(time.time()*1000))
        }
                
        if custom_id is not False:
            params['newClientOrderId'] = custom_id
                
        if amount is not None:
            params['quantity'] = format(self.toValidQuantity(symbol, amount, round_up_amount), 'f')
        elif quote_amount is not None:
            params['quoteOrderQty'] = format(self.toValidQuantity(symbol, 
                                                                quote_amount, round_up_amount), 'f')

        if verbose:
            pprint(params)

        return self.placeOrder(params, test)

    def placeTakeProfitLimitOrder(self, symbol:str, price, stop_price, side:str, amount, quote_amount=None, 
    test:bool=False, round_up_price=False, round_up_amount=False, custom_id=False, verbose=False):
        """ Places a TAKE_PROFIT_LIMIT order for amount of symbol at price. """
        
        if verbose:
            print("Placing TakeProfitLimitOrder with params:")
            
        params = {
            'symbol': symbol.upper(),
            'side': side,
            'type': Binance.ORDER_TYPE_TAKE_PROFIT_LIMIT,
            'recvWindow': 5000,
            'timeInForce': 'GTC',
            'price': format(self.toValidPrice(symbol, price, round_up_price), 'f'),
            'stopPrice': format(self.toValidPrice(symbol, stop_price, round_up_price), 'f'),
            'timestamp': int(round(time.time()*1000))
        }
                
        if custom_id is not False:
            params['newClientOrderId'] = custom_id
                
        if amount is not None:
            params['quantity'] = format(self.toValidQuantity(symbol, 
                                                                        amount, round_up_amount), 'f')
        elif quote_amount is not None:
            params['quoteOrderQty'] = format(self.toValidQuantity(symbol, 
                                                                quote_amount, round_up_amount), 'f')

        if verbose:
            pprint(params)

        return self.placeOrder(params, test)

    def cancelOrder(self, symbol, order_id, is_custom_id=False):
        """ Cancels order given order id """
        params = {
            'symbol': symbol.upper(),
            'recvWindow': 5000,
            'timestamp': int(round(time.time()*1000)) 
        }
        if is_custom_id:
            params['origClientOrderId'] = order_id
        else:
            params['orderId'] = order_id

        self._signRequest(params)
        url = Binance.BASE_URL + Binance.ENDPOINTS['order']
        return self._delete(url, params, self.headers)

    def getOrder(self, symbol, order_id, is_custom_id=True):
        """ Gets order info given order id """
        params = {
            'symbol': symbol.upper(),
            'recvWindow': 5000,
            'timestamp': int(round(time.time()*1000)) 
        }
        if is_custom_id:
            params['origClientOrderId'] = order_id
        else:
            params['orderId'] = order_id
        print(params)
        self._signRequest(params)
        url = Binance.BASE_URL + Binance.ENDPOINTS['order']
        return self._get(url, params, self.headers)

    def getAllOrders(self, symbol, limit=500):
        """ Gets order info given order id """
        params = {
            'symbol': symbol.upper(),
            'limit': limit,
            'timestamp': int(round(time.time()*1000)) 
        }
        self._signRequest(params)
        url = Binance.BASE_URL + Binance.ENDPOINTS['allOrders']
        return self._get(url, params, self.headers)

    @classmethod
    def isValidResponse(cls, response:dict):
        """ 
        Checks whether response received from exchange is valid
        Returns 
        --
        True if valid, False otherwise
        """
        return not response.__contains__('code')

    @classmethod
    def floatToString(cls, f:float):
        ''' Converts the given float to a string,
        without resorting to the scientific notation '''

        ctx = Context()
        ctx.prec = 32
        d1 = ctx.create_decimal(repr(f))
        return format(d1, 'f')

    @classmethod
    def _get10Factor(cls, num):
        """ Returns the power of 10 with which this number needs to be multiplied 
        so that it's between 1 (inclusive) and 10 (exclusive)
        
        _get10Factor(0.00000164763) = 6
        _get10Factor(1600623.3) = -6
        """
        p = 0
        for i in range(-20, 20):
            if num == num % 10**i:
                p = -(i - 1)
                break
        return p

    @classmethod
    def _round_down_decimals(cls, number, decimals):
        factor = 10 ** decimals
        return math.floor(Decimal(number) * Decimal(factor)) / Decimal(factor)

    def toValidPrice(self, symbol, desired_price, round_up:bool=False) \
        -> Decimal():
        """ Returns the minimum quantity of a symbol we can buy, 
        closest to desiredPrice """
        
        pr_filter = {}
        symbol = symbol.upper()

        if not self.SYMBOL_DATAS.__contains__(symbol):
            self.updateSymbolsData()
            if not self.SYMBOL_DATAS.__contains__(symbol):
                raise InternalExchangeException(\
                    "Could not find symbol data of "+symbol)

        for fil in self.SYMBOL_DATAS[symbol]["filters"]:
            if fil["filterType"] == "PRICE_FILTER":
                pr_filter = fil
                break
        
        if not pr_filter.keys().__contains__("tickSize"):
            raise InternalExchangeException(\
                "Couldn't find tickSize or PRICE_FILTER in symbol_data.")

        round_off_by = int(Binance._get10Factor((float(pr_filter["tickSize"]))))
        number = round(Decimal(desired_price), round_off_by)
        if round_up:
            number = number + Decimal(pr_filter["tickSize"])

        return number

    def toValidQuantity(self, symbol, desired_quantity, round_up:bool=False) \
        -> Decimal():
        """ Returns the minimum quantity of a symbol we can buy,
        closest to desiredPrice """
        
        lot_filter = {}

        symbol = symbol.upper()
        # Check whether SD exists
        if not self.SYMBOL_DATAS.__contains__(symbol):
            self.updateSymbolsData()
            if not self.SYMBOL_DATAS.__contains__(symbol):
                raise InternalExchangeException(\
                    "Could not find symbol data of "+symbol)

        # Check the LOT SIZE filter on SD
        for fil in self.SYMBOL_DATAS[symbol]["filters"]:
            if fil["filterType"] == "LOT_SIZE":
                lot_filter = fil
                break
        
        if not lot_filter.keys().__contains__("stepSize"):
            raise InternalExchangeException(\
                "Couldn't find stepSize or PRICE_FILTER in symbol_data.")

        decimals = int(self._get10Factor((float(lot_filter["stepSize"]))))
        quantity = self._round_down_decimals(number = desired_quantity, decimals = decimals)
        if round_up:
            quantity = quantity + Decimal(lot_filter["stepSize"])

        return quantity

    def updateSQLOrderModel(self, order, new_order_response, bot):
        """ Updates an order based on it's state on the exchange given by 
        new_order_response. Should be part of the exchange interface  """

        if order.is_test:
            if order.side == 'BUY':
                order.entry_price = order.price
            
        if not order.is_test:
            if new_order_response.__contains__('transactTime'):
                order.timestamp = datetime.fromtimestamp(
                    new_order_response['transactTime']/1000)
            else:
                order.timestamp = datetime.fromtimestamp(
                    new_order_response['time']/1000)
            # pprint(new_order_response)
            order.price = new_order_response['price']
            order.original_quantity =  Decimal(new_order_response['origQty'])
            order.executed_quantity =  Decimal(new_order_response['executedQty'])
            order.status = new_order_response['status']
            order.side = new_order_response['side']
            order.order_type = new_order_response['type']
            if order.side == 'BUY':
                order.entry_price = new_order_response['price']

        return order



from decimal import Decimal
from uuid import uuid4
import time
import math
from pprint import pprint
from pyjuque.Engine.Models import TABotModel as Bot, PairModel as Pair, OrderModel as Order
from pyjuque.Engine.OrderManager import placeNewOrder, simulateOrderInfo, cancelOrder
from pyjuque.Exchanges.Base.Exceptions import InvalidCredentialsException, \
    InternalExchangeException, ExchangeConnectionException
from traceback import print_exc
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
import sys

class BotController:

    def __init__(self, session, bot_model, exchange, strategy, timeframe = '5m', status_printer = None, logger_on = False):
        self.bot_model = bot_model
        self.session = session
        self.exchange = exchange
        self.strategy = strategy
        self.test_mode = bot_model.test_run
        self.kline_interval = timeframe
        self.status_printer = status_printer
        self.logger_on = logger_on


    def executeBot(self):
        """ The main execution loop of the bot """
        if self.status_printer != None:
            self.status_printer.start()
        # Step 1: Retreive all pairs for a particular bot
        self.log("Getting active pairs:", should_print=False)
        active_pairs = self.bot_model.getActivePairs(self.session)
        self.log("Number of active_pairs: {}".format(
                len(active_pairs)), should_print=False)
        # Step 2 For each pair:
        #	Retreive current market data
        # 	Compute indicators & check if strategy
        #		If strategy fulfilled, palce order & save order in DB
        self.log("Checking signals on pairs...", should_print=False)
        for pair in active_pairs:
            self.log("Checking signal on {}".format(pair.symbol), should_print=False)
            self.tryEntryOrder(pair)
        # Step 3: Retreive all open orders on the bot
        self.log("Getting open orders:", should_print=False)
        open_orders = self.bot_model.getOpenOrders(self.session)
        self.log("Number of open orders: {}".format(len(open_orders)), should_print=False)
        # Step 4: For Each order that was already placed by the bot
        # and was not filled before, check status:
        #       IF Filled 
        #           -> If entry order, place exit order
        #           -> If exit order, success (take profit), 
        #           or failure (stop loss): Resume trading!
        self.log("Checking orders state...", should_print=False)
        for order in open_orders:
            self.updateOpenOrder(order)
        
        self.log("Executed the bot loop. Now waiting...", should_print=False)


    def checkEntryStrategy(self, symbol):
        """ Default function that checks up the entry strategy. 
        Can be overridden for custom strategies.
        """
        try:
            df = self.exchange.getOHLCV(symbol, self.kline_interval, self.strategy.minimum_period)
        except Exception as e:
            self.log('Error getting data from the exchange for {}:'.format(symbol))
            self.logError(sys.exc_info())
            return False, None
        try:
            self.strategy.setUp(df)
            entry_signal = self.strategy.checkLongSignal(len(df) - 1)
            last_price = df.iloc[-1]['close']
        except Exception as e:
            self.log('Error computing indicators for {}:'.format(symbol))
            self.logError(sys.exc_info())
            return False, None
        return entry_signal, last_price


    def tryEntryOrder(self, pair):
        """
        Gets the latest market data and runs the strategy off of it. 
        If strategy returns entry signal, entry is made.
        """
        entry_signal, last_price = self.checkEntryStrategy(pair.symbol)
        if entry_signal:
            self.log('Got entry signal on {}.'.format(pair.symbol))
            side = 'buy'
            quote_qty = self.bot_model.starting_balance \
                * Decimal(self.bot_model.entry_settings.initial_entry_allocation) / 100
            desired_price = None
            if self.bot_model.entry_settings.signal_distance == 0:
                order_type = 'market'
            else:
                desired_price = Decimal(last_price) \
                    * Decimal((100 - self.bot_model.entry_settings.signal_distance) / 100 )
                order_type = 'limit'
            quantity = quote_qty / desired_price
            if self.bot_model.current_balance > 0:
                self.placeOrder(
                    symbol=pair.symbol, 
                    pair=pair, 
                    side=side, 
                    price=desired_price,  
                    quantity=quantity, 
                    order_type=order_type, 
                    is_entry=True)
            else:
                self.log('Not placing order on {} due to insuficient balance.'.format(pair.symbol))
            self.session.commit()


    def updateOpenOrder(self, order):
        """
        Checks the status of an open order and updates it
        """
        pair = self.bot_model.getPair(self.session, symbol=order.symbol)
        self.log('Checking open order on {}.'.format(order.symbol), should_print=False)
        # get info of order from exchange
        if not self.test_mode: 
            try:
                exchange_order_info = self.exchange.getOrder(order.symbol, order.id, is_custom_id=True)
            except Exception as e:
                self.log('Error getting data from the exchange for'
                    ' updating open order on {}.'.format(order.symbol))
                self.logError(sys.exc_info())
                return
        else:
            try:
                exchange_order_info = simulateOrderInfo(self.exchange, order, self.kline_interval)
            except Exception as e:
                self.log('Error simulating open order on {}.'.format(order.symbol))
                self.logError(sys.exc_info())
                return
        self.exchange.updateSQLOrderModel(order, exchange_order_info, None)
        self.session.commit()
        # update order params.
        # order.side = exchange_order_info['side']
        # order.status = exchange_order_info['status']
        # order.executed_quantity = exchange_order_info['filled']
        if (order.side == 'buy'):
            # Order has been canceled by the user
            if (order.status == 'canceled'):
                self.log('CANCELED BUY order on {}, exiting remaining quantity: {} at market price.'.format(order.symbol, order.executed_quantity))
                self.instantlyExitOrder(order, pair)
                self.processClosedPosition(order, pair)
            # buy order was filled, place exit order.
            if (order.status == 'closed'):
                self.log('FILLED BUY order on {}, try exit.'.format(order.symbol))
                self.tryExitOrder(order, pair)
            # buy order has been accepted by engine
            if (order.status == 'open'):
                self.log('OPEN BUY order on {}, UPDATING.'.format(order.symbol))
                self.updateOpenBuyOrder(order, pair)
            # buy order was rejected, not processed by engine
            if (order.status == 'rejected'):
                self.log('REJECTED BUY order on {}.'.format(order.symbol))
                self.processClosedPosition(order, pair)
            # buy order expired, i.e. FOK orders with 
            # no fill or due to maintenance by exchange.
            if (order.status == 'expired'):
                self.log('EXPIRED BUY order on {}, sell what was bought.'.format(order.symbol))
                self.instantlyExitOrder(order, pair)
                self.processClosedPosition(order, pair)
        # sell order 
        elif (order.side == 'sell'):
            # was cancelled by user.
            if (order.status == 'canceled'):
                # query buy order to again place a sell order for that buy order.
                # self.log('CANCELED SELL order on {}.'.format(order.symbol))
                # original_buy_order = self.reviveOriginalBuyOrder(order)
                # self.tryExitOrder(original_buy_order, pair)
                self.log('CANCELED SELL order on {}'.format(order.symbol))
                # self.instantlyExitOrder(order, pair)
                self.processClosedPosition(order, pair)
            # sell order was filled
            if (order.status == 'closed'):
                self.log('FILLED SELL order on {}.'.format(order.symbol))
                self.processClosedPosition(order, pair)
            # sell order was accepted by engine of exchange
            if (order.status == 'open'):
                self.log('OPEN SELL order on {}, UPDATING.'.format(order.symbol))
                try:
                    self.updateOpenSellOrder(order, pair)
                except:
                    print('Error trying to update open sell order on {} with id {}'.format(order.symbol, order.id))
            # sell order was rejected by engine of exchange
            if (order.status == 'rejected'):
                self.log('REJECTED SELL order on {}.'.format(order.symbol))
                original_buy_order = self.reviveOriginalBuyOrder(order)
                self.tryExitOrder(original_buy_order, pair)
            # sell order expired, i.e. due to FOK orders or partially filled market orders
            if (order.status == 'expired'):
                self.log('EXPIRED SELL order on {}.'.format(order.symbol))
                original_buy_order = self.reviveOriginalBuyOrder(order)
                self.tryExitOrder(original_buy_order, pair)
        self.session.commit()


    def reviveOriginalBuyOrder(self, order):
        """ If sell order was cancelled due to some reason, revive buy order and look to exit again. """
        original_buy_order = self.session.query(Order).get(order.id)
        original_buy_order.is_closed = False
        original_buy_order.executed_quantity = order.original_quantity - order.executed_quantity
        return original_buy_order


    def updateOpenBuyOrder(self, order, pair):
        """
        Looks to close an open buy order based on strategy specified 
        in entry settings. For now you can only close orders based on 
        time passed.
        """
        # TODO: place a sell order to exit if holding any 
        # asset bought through this order.
        if (time.time() * 1000 - (order.timestamp).timestamp()) \
            > self.bot_model.entry_settings.open_buy_order_time_out:
            if not self.test_mode:
                try:
                    order_result = cancelOrder(self.exchange, order)
                    if order.executed_quantity > 0:
                        self.instantlyExitOrder(order, pair)
                        self.bot_model.current_balance += \
                            order.executed_quantity * order.price
                    else:
                        self.bot_model.current_balance += \
                            order.original_quantity * order.price
                except Exception as e:
                    self.log('cancelOrder() failed on {}'.format(order.symbol))
                    self.log(e)
                    return
                # try:
                #     order_result = self.exchange.getOrder(order.symbol, order.id, is_custom_id=True)
                # except Exception as e:
                #     self.log('getting order failed after canceling order on {}'.format(order.symbol))
                #     self.log(e)
                #     return
            self.log('Canceled BUY order on {} due to timeout.'.format(order.symbol))
            # self.exchange.updateSQLOrderModel(order, order_result, None)
            self.processClosedPosition(order, pair)


    def updateOpenSellOrder(self, order, pair):
        """
        Update open sell order based on given params in exit settings.
        For now only supports placing stop loss market order when 
        position goes against us (basically simple oco order variant)
        """
        try:
            candlestick_data = self.exchange.getOHLCV(order.symbol, self.kline_interval)
        except Exception:
            self.log('Error getting data from the exchange '
                'for updating open sell order on {}:'.format(pair.symbol))
            self.logError(sys.exc_info())
            return
        current_price = Decimal(candlestick_data.iloc[-1]['close'])
        stop_loss_value = self.bot_model.exit_settings.stop_loss_value

        exit_signal, last_price = self.checkExitStrategy(order.symbol)
        side = 'sell'
        order_type = 'market'
        if (stop_loss_value != None and (order.stop_price != None and order.stop_price > current_price)) or \
            (self.bot_model.exit_settings.exit_on_signal and exit_signal):
            quantity = self.computeMatchingOrderQuantity(order)
            if not self.test_mode:
                try:
                    cancelOrder(self.exchange, order)
                    if order.executed_quantity > 0:
                        self.bot_model.current_balance += \
                            order.executed_quantity * order.price
                except Exception as e:
                    self.log('cancelOrder() failed on {}'.format(order.symbol))
                    self.log(e)
                    return
                # try:
                #     order_result = self.exchange.getOrder(order.symbol, order.id, is_custom_id=True)
                # except Exception as e:
                #     self.log('getting order failed after canceling order')
                #     self.log(e)
                #     return
            else:
                order_result = dict(status='canceled', filled=0)

            if (stop_loss_value != None and (order.stop_price > current_price)):
                self.log('Canceling existing sell order and placing market sell on {} due to stop loss.'.format(order.symbol))
            elif (self.bot_model.exit_settings.exit_on_signal and exit_signal):
                self.log('Canceling existing sell order and placing market sell on {} due to exit signal.'.format(order.symbol))
            else:
                self.log('Canceling existing sell order and placing market sell on {} ...'.format(order.symbol))
            # pprint(order_result)
            order.status = 'canceled' #order_result['status']
            # order.executed_quantity = order_result['filled']
            order.is_closed = True
            self.placeOrder(
                symbol=order.symbol,
                pair=pair,
                order=order,
                quantity=quantity,
                order_type=order_type,
                side=side)


    def checkExitStrategy(self, symbol):
        """  Default function that checks the exit strategy. 
        Can be overwritten for custom strategies.
        """
        try:
            df = self.exchange.getOHLCV(symbol, self.kline_interval, self.strategy.minimum_period)
        except Exception:
            self.log('Error getting data from the exchange for {}:'.format(symbol))
            self.logError(sys.exc_info())
            return False, None
        try:
            self.strategy.setUp(df)
            exit_signal = self.strategy.checkShortSignal(len(df) - 1)
            last_price = df.iloc[-1]['close']
        except Exception as e:
            self.log('Error computing indicators for {}:'.format(symbol))
            self.logError(sys.exc_info())
            return False, None
        return exit_signal, last_price


    def tryExitOrder(self, order, pair):
        """ If strategy returns exit signal look to place exit order. """
        # Calculates quantity of order. 
        # Takes in to account partially filled orders.
        quantity = self.computeMatchingOrderQuantity(order)
        if quantity > 0:
            price = None
            stop_price = None
            exit_signal, last_price = self.checkExitStrategy(order.symbol)
            place_trade = False
            if self.bot_model.exit_settings.exit_on_signal and exit_signal:
                order_type = 'market'
                side = 'sell'
                place_trade = True
            else:
                last_price = order.price
                stop_loss_value = self.bot_model.exit_settings.stop_loss_value 
                profit_target = self.bot_model.exit_settings.profit_target
                if profit_target is not None:
                    price = Decimal(last_price) * Decimal((100 + profit_target) / 100)
                    order_type = 'limit'
                    side = 'sell'
                    place_trade = True
                    if stop_loss_value is not None:
                        stop_price = Decimal(last_price) * Decimal((100 - stop_loss_value) / 100)
                elif stop_loss_value is not None:
                    price = Decimal(last_price) * Decimal((100 - stop_loss_value) / 100)
                    order_type = 'stop_loss'
                    side = 'sell'
                    place_trade = True
            # If trade can go through
            if place_trade:
                self.placeOrder(
                    symbol=order.symbol, 
                    pair=pair, 
                    order=order, 
                    price=price,
                    entry_price=order.price, 
                    stop_price=stop_price,
                    quantity=quantity, 
                    side=side, 
                    order_type=order_type)


    def instantlyExitOrder(self, order, pair):
        """ If strategy returns exit signal look to place exit order. """
        quantity = self.computeMatchingOrderQuantity(order)
        order_type = 'market'
        side = 'buy'
        if order.side == 'buy':
            side = 'sell'
        if quantity > 0:
            self.placeOrder(
                symbol=order.symbol, 
                pair=pair, 
                order=order,
                quantity=quantity, 
                side=side, 
                order_type=order_type)


    def processClosedPosition(self, order, pair):
        """ Update status of closed position """
        order.is_closed = True
        pair.active = True
        pair.current_order_id = None
        if order.side == 'sell':
            self.bot_model.current_balance += \
                order.executed_quantity * order.price


    def placeOrder(self, symbol, pair, order=None, **order_params):
        """ Create Order model and place order to exchange. """
        order_params['bot_id'] = self.bot_model.id
        new_order_model = None
        try:
            new_order_model = placeNewOrder(
                self.exchange, symbol, pair, order, self.bot_model.test_run, order_params)
        except Exception as e:
            self.log('Error placing order for {}:'.format(symbol))
            self.logError(sys.exc_info())
        if new_order_model != None:
            self.syncModels(pair, order, new_order_model)
            self.session.add(new_order_model)
            self.log('{} Order was placed succesfully on {}'.format(
                new_order_model.side, symbol))


    def syncModels(self, pair, order, new_order_model):
        """ Sync pairs and orders to new status """
        if order is not None:
            new_order_model.matched_order_id = order.id
            order.is_closed = True
            order.matched_order_id = new_order_model.id
            pair.active = False
            pair.current_order_id = new_order_model.id
        if order is None:
            self.bot_model.current_balance -= \
                Decimal(self.bot_model.entry_settings.initial_entry_allocation) \
                    * self.bot_model.starting_balance / Decimal(100)
            pair.active = False
            pair.current_order_id = new_order_model.id

    # Keep Bot Balance Updated
    # Exit on Signal
    # Activate Simulation Mode

    def computeMatchingOrderQuantity(self, order):
        """ Compute exit quantity so that also partially 
        filled orders can be handled."""
        if order.side == 'buy':
            exit_quantity = order.executed_quantity
        elif order.side == 'sell':
            exit_quantity = order.original_quantity \
                - order.executed_quantity
        return exit_quantity


    def log(self, message, should_print=True, force=True):
        if self.status_printer != None:
            self.status_printer.stop()
            if should_print:
                logger.info(message)
            else:
                self.status_printer.text = message
            self.status_printer.start()
        elif self.logger_on or force:
            logger.info(message)


    def logError(self, message):
        if self.status_printer != None:
            self.status_printer.stop()
            print(message)
            self.status_printer.start()
        else:
            print(message)


from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
	name='pyjuque',
	packages=find_packages(),
	version='0.1.1.7',
	description="'Juju' Quant Engine for Python",
	long_description=long_description,
	long_description_content_type="text/markdown",
	author='Tudor Barbulescu',
	author_email='hello@tudorbarbulescu.com',
	license='MIT',
    url='https://github.com/tudorelu/pyjuque',
	install_requires=['yaspin==1.2.0', 'ccxt==1.40.74', 'pyyaml==5.4', 'numpy==1.19.4', 'pandas==1.1.4', 'plotly==4.12.0', 'python-dotenv==0.15.0' ,'requests==2.25.0', 'SQLAlchemy==1.3.18', 'websocket==0.2.1', 'websocket-client==0.57.0'],
	setup_requires=['nose2==0.9.2'],
	tests_require=['alchemy-mock==0.4.3', 'coverage==5.1', 'freezegun==0.3.15', 'mock==4.0.2', 'nose2==0.9.2']
)
